# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from future import standard_library

standard_library.install_aliases()
from builtins import range
import urllib.request, urllib.parse, urllib.error
import json
from traceback import format_exc

import splunktalib.common.util as scu
import splunktalib.common.log as log

from httplib2 import socks, ProxyInfo, Http


def splunkd_request(
    splunkd_uri, session_key, method="GET", headers=None, data=None, timeout=30, retry=1
):
    """
    :return: httplib2.Response and content
    """

    headers = headers if headers is not None else {}
    headers["Authorization"] = "Splunk {0}".format(session_key)
    content_type = headers.get("Content-Type")
    if not content_type:
        content_type = headers.get("content-type")

    if not content_type:
        content_type = "application/x-www-form-urlencoded"
        headers["Content-Type"] = content_type

    if data is not None:
        if content_type == "application/json":
            data = json.dumps(data)
        else:
            data = urllib.parse.urlencode(data)

    http = Http(timeout=timeout, disable_ssl_certificate_validation=True)
    msg_temp = "Failed to send rest request=%s, errcode=%s, reason=%s"
    resp, content = None, None
    for _ in range(retry):
        try:
            resp, content = http.request(
                splunkd_uri, method=method, headers=headers, body=data
            )
        except Exception:
            log.logger.error(msg_temp, splunkd_uri, "unknown", format_exc())
        else:
            if resp.status not in (200, 201):
                if not (method == "GET" and resp.status == 404):
                    log.logger.debug(
                        msg_temp, splunkd_uri, resp.status, code_to_msg(resp, content)
                    )
            else:
                return resp, content
    else:
        return resp, content


def code_to_msg(resp, content):
    code_msg_tbl = {
        400: "Request error. reason={}".format(content),
        401: "Authentication failure, invalid access credentials.",
        402: "In-use license disables this feature.",
        403: "Insufficient permission.",
        404: "Requested endpoint does not exist.",
        409: "Invalid operation for this endpoint. reason={}".format(content),
        500: "Unspecified internal server error. reason={}".format(content),
        503: (
            "Feature is disabled in the configuration file. "
            "reason={}".format(content)
        ),
    }

    return code_msg_tbl.get(resp.status, content)


def build_http_connection(config, timeout=120, disable_ssl_validation=False):
    """
    :config: dict like, proxy and account information are in the following
             format {
                 "username": xx,
                 "password": yy,
                 "proxy_url": zz,
                 "proxy_port": aa,
                 "proxy_username": bb,
                 "proxy_password": cc,
                 "proxy_type": http,http_no_tunnel,sock4,sock5,
                 "proxy_rdns": 0 or 1,
             }
    :return: Http2.Http object
    """

    proxy_type_to_code = {
        "http": socks.PROXY_TYPE_HTTP,
        "http_no_tunnel": socks.PROXY_TYPE_HTTP_NO_TUNNEL,
        "socks4": socks.PROXY_TYPE_SOCKS4,
        "socks5": socks.PROXY_TYPE_SOCKS5,
    }
    if config.get("proxy_type") in proxy_type_to_code:
        proxy_type = proxy_type_to_code[config["proxy_type"]]
    else:
        proxy_type = socks.PROXY_TYPE_HTTP

    rdns = scu.is_true(config.get("proxy_rdns"))

    proxy_info = None
    if config.get("proxy_url") and config.get("proxy_port"):
        if config.get("proxy_username") and config.get("proxy_password"):
            proxy_info = ProxyInfo(
                proxy_type=proxy_type,
                proxy_host=config["proxy_url"],
                proxy_port=int(config["proxy_port"]),
                proxy_user=config["proxy_username"],
                proxy_pass=config["proxy_password"],
                proxy_rdns=rdns,
            )
        else:
            proxy_info = ProxyInfo(
                proxy_type=proxy_type,
                proxy_host=config["proxy_url"],
                proxy_port=int(config["proxy_port"]),
                proxy_rdns=rdns,
            )
    if proxy_info:
        http = Http(
            proxy_info=proxy_info,
            timeout=timeout,
            disable_ssl_certificate_validation=disable_ssl_validation,
        )
    else:
        http = Http(
            timeout=timeout, disable_ssl_certificate_validation=disable_ssl_validation
        )

    if config.get("username") and config.get("password"):
        http.add_credentials(config["username"], config["password"])
    return http



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from future import standard_library

standard_library.install_aliases()
from six import string_types
from builtins import object
import queue
import multiprocessing
import threading
import sys
from collections import Iterable
from splunktalib.common import log


class EventWriter(object):
    def __init__(self, process_safe=False):
        if process_safe:
            self._mgr = multiprocessing.Manager()
            self._event_queue = self._mgr.Queue(1000)
        else:
            self._event_queue = queue.Queue(1000)
        self._event_writer = threading.Thread(target=self._do_write_events)
        self._event_writer.daemon = True
        self._started = False
        self._exception = False

    def start(self):
        if self._started:
            return
        self._started = True

        self._event_writer.start()
        log.logger.info("Event writer started.")

    def tear_down(self):
        if not self._started:
            return
        self._started = False

        self._event_queue.put(None)
        self._event_writer.join()
        log.logger.info("Event writer stopped.")

    def isopen(self):
        return self._started and (not self._exception)

    def write_events(self, events):
        if not self.isopen():
            return False
        if events is None:
            return True
        self._event_queue.put(events)
        return True

    def _do_write_events(self):
        event_queue = self._event_queue
        write = sys.stdout.write
        got_shutdown_signal = False

        while 1:
            try:
                event = event_queue.get(timeout=3)
                if event is not None:
                    if isinstance(event, string_types):
                        write(event.encode("utf-8"))
                    elif isinstance(event, Iterable):
                        for evt in event:
                            write(evt.encode("utf-8"))
                else:
                    log.logger.info("Event writer got tear down signal")
                    got_shutdown_signal = True
            except queue.Empty:
                # We need drain the queue before shutdown
                # timeout means empty for now
                if got_shutdown_signal:
                    log.logger.info("Event writer is going to exit...")
                    break
                else:
                    continue
            except Exception:
                log.logger.exception(
                    "EventWriter encounter exception which may"
                    "cause data loss, queue leftsize={"
                    "}".format(event_queue.qsize())
                )
                self._exception = True
                break

        log.logger.info(
            "Event writer stopped, queue leftsize={}".format(event_queue.qsize())
        )


class EventWriterWithCheckpoint(EventWriter):
    def _do_write_events(self):
        event_queue = self._event_queue
        write = sys.stdout.write
        got_shutdown_signal = False

        while 1:
            try:
                event = event_queue.get(timeout=3)
                if event is not None:
                    # event is a tuple which consists of events and checkpoint
                    # information: (events, checkpoint_tuple)
                    # checkpoint_tuple includes the checkpoint manager obj, key
                    # and value of checkpoint: (ckpt_mgr_obj, key, state)
                    events = event[0]
                    ckpt_tuple = event[1]
                    if isinstance(events, string_types):
                        write(events.encode("utf-8"))
                    elif isinstance(events, Iterable):
                        for evt in events:
                            write(evt.encode("utf-8"))

                    # Update checkpoint after events are sent to stdout to avoid
                    # data loss.
                    if ckpt_tuple and ckpt_tuple[2]:
                        ckpt_tuple[0].update_state(ckpt_tuple[1], ckpt_tuple[2])

                    # Close the checkpoint obj to flush the data in cache to
                    # disk to aviod data duplication if it is teared down
                    if not self._started:
                        ckpt_tuple[0].close()
                else:
                    log.logger.info("Event writer got tear down signal")
                    got_shutdown_signal = True

            except queue.Empty:
                # We need drain the queue before shutdown
                # timeout means empty for now
                if got_shutdown_signal:
                    log.logger.info("Event writer is going to exit...")
                    break
                else:
                    continue
            except Exception:
                log.logger.exception(
                    "EventWriter encounter exception which may"
                    "cause data loss, queue leftsize={"
                    "}".format(event_queue.qsize())
                )
                self._exception = True
                break

        log.logger.info(
            "Event writer stopped, queue leftsize={}".format(event_queue.qsize())
        )



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright (C) 2005-2019 Splunk Inc. All Rights Reserved.
"""

from future import standard_library

standard_library.install_aliases()
from builtins import str
import os
import os.path as op
import datetime
import sys
import gc
import urllib.request, urllib.parse, urllib.error


def handle_tear_down_signals(callback):
    import signal

    signal.signal(signal.SIGTERM, callback)
    signal.signal(signal.SIGINT, callback)

    if os.name == "nt":
        signal.signal(signal.SIGBREAK, callback)


def datetime_to_seconds(dt):
    epoch_time = datetime.datetime.utcfromtimestamp(0)
    return (dt - epoch_time).total_seconds()


def is_true(val):
    value = str(val).strip().upper()
    if value in ("1", "TRUE", "T", "Y", "YES"):
        return True
    return False


def is_false(val):
    value = str(val).strip().upper()
    if value in ("0", "FALSE", "F", "N", "NO", "NONE", ""):
        return True
    return False


def remove_http_proxy_env_vars():
    for k in ("http_proxy", "https_proxy"):
        if k in os.environ:
            del os.environ[k]
        elif k.upper() in os.environ:
            del os.environ[k.upper()]


def get_appname_from_path(absolute_path):
    absolute_path = op.normpath(absolute_path)
    parts = absolute_path.split(os.path.sep)
    parts.reverse()
    for key in ("apps", "slave-apps", "master-apps"):
        try:
            idx = parts.index(key)
        except ValueError:
            continue
        else:
            try:
                if parts[idx + 1] == "etc":
                    return parts[idx - 1]
            except IndexError:
                pass
            continue
    return "-"


def escape_cdata(data):
    data = data.encode("utf-8", errors="xmlcharrefreplace").decode("utf-8")
    data = data.replace("]]>", "]]&gt;")
    if data.endswith("]"):
        data = data[:-1] + "%5D"
    return data


def extract_datainput_name(stanza_name):
    """
    stansa_name: string like aws_s3://my_s3_data_input
    """

    sep = "://"
    try:
        idx = stanza_name.index(sep)
    except ValueError:
        return stanza_name

    return stanza_name[idx + len(sep) :]


def escape_json_control_chars(json_str):
    control_chars = ((r"\n", "\\\\n"), (r"\r", "\\\\r"), (r"\r\n", "\\\\r\\\\n"))
    for ch, replace in control_chars:
        json_str = json_str.replace(ch, replace)
    return json_str


def disable_stdout_buffer():
    os.environ["PYTHONUNBUFFERED"] = "1"
    sys.stdout = os.fdopen(sys.stdout.fileno(), "wb", 0)
    gc.garbage.append(sys.stdout)


def format_stanza_name(name):
    return urllib.parse.quote(name.encode("utf-8"), "")



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

import splunktalib.rest as rest
from splunktalib.common import log


class ConfRequestException(Exception):
    pass


class ConfNotExistsException(ConfRequestException):
    pass


class ConfExistsException(ConfRequestException):
    pass


def content_request(uri, session_key, method, payload, err_msg):
    """
    :return: response content if successful otherwise raise
    ConfRequestException
    """

    resp, content = rest.splunkd_request(
        uri, session_key, method, data=payload, retry=3
    )
    if resp is None and content is None:
        return None

    if resp.status >= 200 and resp.status <= 204:
        return content
    else:
        msg = "%s, status=%s, reason=%s, detail=%s" % (
            err_msg,
            resp.status,
            resp.reason,
            content.decode("utf-8"),
        )

        if not (method == "GET" and resp.status == 404):
            log.logger.error(msg)

        if resp.status == 404:
            raise ConfNotExistsException(msg)
        if resp.status == 409:
            raise ConfExistsException(msg)
        else:
            if content and "already exists" in content:
                raise ConfExistsException(msg)
            raise ConfRequestException(msg)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright (C) 2005-2019 Splunk Inc. All Rights Reserved.

log utility for TA
"""

from builtins import object
import logging
import logging.handlers as handlers
import os.path as op
from splunktalib.splunk_platform import make_splunkhome_path
import splunktalib.common.util as cutil
from splunktalib.common.pattern import singleton

import time

logging.Formatter.converter = time.gmtime


def log_enter_exit(logger):
    """
    Log decorator to log function enter and exit
    """

    def log_decorator(func):
        def wrapper(*args, **kwargs):
            logger.debug("{} entered.".format(func.__name__))
            result = func(*args, **kwargs)
            logger.debug("{} exited.".format(func.__name__))
            return result

        return wrapper

    return log_decorator


@singleton
class Logs(object):
    def __init__(self, namespace=None, default_level=logging.INFO):
        self._loggers = {}
        self._default_level = default_level
        if namespace is None:
            namespace = cutil.get_appname_from_path(op.abspath(__file__))

        if namespace:
            namespace = namespace.lower()
        self._namespace = namespace

    def get_logger(self, name, level=None, maxBytes=25000000, backupCount=5):
        """
        Set up a default logger.

        :param name: The log file name.
        :param level: The logging level.
        :param maxBytes: The maximum log file size before rollover.
        :param backupCount: The number of log files to retain.
        """

        # Strip ".py" from the log file name if auto-generated by a script.
        if level is None:
            level = self._default_level

        name = self._get_log_name(name)
        if name in self._loggers:
            return self._loggers[name]

        logfile = make_splunkhome_path(["var", "log", "splunk", name])
        logger = logging.getLogger(name)

        handler_exists = any(
            [True for h in logger.handlers if h.baseFilename == logfile]
        )
        if not handler_exists:
            file_handler = handlers.RotatingFileHandler(
                logfile, mode="a", maxBytes=maxBytes, backupCount=backupCount
            )

            formatter = logging.Formatter(
                "%(asctime)s +0000 log_level=%(levelname)s, pid=%(process)d, tid=%(threadName)s, "
                "file=%(filename)s, func_name=%(funcName)s, code_line_no=%(lineno)d | %(message)s"
            )
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
            logger.setLevel(level)
            logger.propagate = False

        self._loggers[name] = logger
        return logger

    def set_level(self, level, name=None):
        """
        Change the log level of the logging

        :param level: the level of the logging to be setLevel
        :param name: the name of the logging to set, in case it is not set,
                     all the loggers will be affected
        """

        if name is not None:
            name = self._get_log_name(name)
            logger = self._loggers.get(name)
            if logger is not None:
                logger.setLevel(level)
        else:
            self._default_level = level
            for logger in self._loggers.values():
                logger.setLevel(level)

    def _get_log_name(self, name):
        if name.endswith(".py"):
            name = name.replace(".py", "")

        if self._namespace:
            name = "{}_{}.log".format(self._namespace, name)
        else:
            name = "{}.log".format(name)
        return name


# Global logger
logger = Logs().get_logger("util")


def reset_logger(name):
    """
    Reset global logger.
    """

    global logger
    logger = Logs().get_logger(name)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import os.path as op
import traceback

from splunktalib.common import log


class FileMonitor(object):
    def __init__(self, callback, files):
        """
        :files: files to be monidtored with full path
        """

        self._callback = callback
        self._files = files

        self.file_mtimes = {file_name: None for file_name in self._files}
        for k in self.file_mtimes:
            if not op.exists(k):
                continue

            try:
                if not op.exists(k):
                    continue
                self.file_mtimes[k] = op.getmtime(k)
            except OSError:
                log.logger.error(
                    "Getmtime for %s, failed: %s", k, traceback.format_exc()
                )

    def __call__(self):
        return self.check_changes()

    def check_changes(self):
        log.logger.debug("Checking files=%s", self._files)
        file_mtimes = self.file_mtimes
        changed_files = []
        for f, last_mtime in file_mtimes.items():
            try:
                if not op.exists(f):
                    continue

                current_mtime = op.getmtime(f)
                if current_mtime != last_mtime:
                    file_mtimes[f] = current_mtime
                    changed_files.append(f)
                    log.logger.info("Detect %s has changed", f)
            except OSError:
                pass

        if changed_files:
            if self._callback:
                self._callback(changed_files)
            return True
        return False



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
This module hanles high level TA configuration related stuff
"""

from builtins import object
import os.path as op
import copy

import splunktalib.common.util as utils
import splunktalib.credentials as cred
import splunktalib.conf_manager.conf_manager as conf
import splunktalib.conf_manager.request as conf_req


class TAConfManager(object):

    encrypted_token = "******"
    reserved_keys = ("userName", "appName")

    def __init__(self, conf_file, splunkd_uri, session_key, appname=None):
        if appname is None:
            appname = utils.get_appname_from_path(op.abspath(__file__))
        self._conf_file = conf.conf_file2name(conf_file)
        self._conf_mgr = conf.ConfManager(splunkd_uri, session_key, app_name=appname)
        self._cred_mgr = cred.CredentialManager(
            splunkd_uri, session_key, app=appname, owner="nobody", realm=appname
        )
        self._keys = None

    def set_appname(self, appname):
        """
        This are cases we need edit/remove/create confs in different app
        context. call this interface to switch app context before manipulate
        the confs in different app context
        """

        self._conf_mgr.set_appname(appname)
        self._cred_mgr.set_appname(appname)

    def _delete_reserved_keys(self, stanza):
        new_stanza = copy.deepcopy(stanza)
        for k in self.reserved_keys:
            if k in new_stanza:
                del new_stanza[k]
        return new_stanza

    def create(self, stanza):
        """
        @stanza: dick like object
        {
        "name": xxx,
        "k1": v1,
        "k2": v2,
        ...
        }
        @return exception if failure
        """

        stanza = self._delete_reserved_keys(stanza)
        encrypted_stanza = self._encrypt(stanza)
        self._conf_mgr.create_stanza(
            self._conf_file, encrypted_stanza["name"], encrypted_stanza
        )

    def update(self, stanza):
        """
        @stanza: dick like object
        {
        "name": xxx,
        "k1": v1,
        "k2": v2,
        ...
        }

        @return: exception if failure
        """

        if not self._conf_mgr.stanza_exist(self._conf_file, stanza["name"]):
            self.create(stanza)
        else:
            stanza = self._delete_reserved_keys(stanza)
            encrypted_stanza = self._encrypt(stanza)
            self._conf_mgr.update_properties(
                self._conf_file, encrypted_stanza["name"], encrypted_stanza
            )

    def delete(self, stanza_name):
        """
        @return: exception if failure
        """

        try:
            stanza = self._conf_mgr.get_stanza(self._conf_file, stanza_name)
        except conf_req.ConfNotExistsException:
            return

        self._delete_creds(stanza)
        self._conf_mgr.delete_stanza(self._conf_file, stanza_name)

    def get(self, stanza_name, return_acl=False):
        """
        @return: dict object if sucess otherwise raise exception
        """

        stanza = self._conf_mgr.get_stanza(
            self._conf_file, stanza_name, ret_metadata=return_acl
        )
        stanza = self._decrypt(stanza)
        stanza["disabled"] = utils.is_true(stanza.get("disabled"))
        return stanza

    def all(self, filter_disabled=False, return_acl=True):
        """
        @return: a dict of dict objects if success
        otherwise exception
        """

        results = {}
        stanzas = self._conf_mgr.all_stanzas(self._conf_file, ret_metadata=return_acl)
        for stanza in stanzas:
            stanza = self._decrypt(stanza)
            stanza["disabled"] = utils.is_true(stanza.get("disabled"))
            if filter_disabled and stanza["disabled"]:
                continue
            results[stanza["name"]] = stanza
        return results

    def reload(self):
        self._conf_mgr.reload_conf(self._conf_file)

    def set_encrypt_keys(self, keys):
        """
        :keys: a list keys of a stanza which need to be encrypted
        for example: ["username", "password"]
        """

        self._keys = keys

    def is_encrypted(self, stanza):
        """
        :stanza: dict object
        return True if the values of encrypt keys equals self.encrypted_token
        otherwise return False
        """

        if self._keys is None:
            return False

        for k in stanza.keys():
            if k in self._keys:
                if stanza.get(k) == self.encrypted_token:
                    return True
        return False

    def _encrypt(self, stanza):
        """
        :stanza: if self._keys are in stanza, encrypt the values of the key
        and then mask the value to self.encrypted_token
        """

        if self._keys is None:
            return stanza

        stanza_to_be_encrypted = {}
        for key in self._keys:
            if key in stanza:
                stanza_to_be_encrypted[key] = stanza[key]

        if stanza_to_be_encrypted:
            self._cred_mgr.update({stanza["name"]: stanza_to_be_encrypted})
            encrypted_stanza = copy.deepcopy(stanza)
            for key in stanza_to_be_encrypted.keys():
                encrypted_stanza[key] = self.encrypted_token
            return encrypted_stanza
        return stanza

    def _decrypt(self, stanza):
        """
        :stanza: if there are keys in self._keys in stanza and if the values of
        the keys are self.encrypted_token, decrypt the value
        """

        if self._keys is None:
            return stanza

        stanza_name = stanza["name"]
        clear_password = None
        for key in self._keys:
            if key in stanza and stanza[key] == self.encrypted_token:
                clear_password = self._cred_mgr.get_clear_password(stanza_name)
                break

        if clear_password:
            for key in self._keys:
                if key in clear_password[stanza_name]:
                    stanza[key] = clear_password[stanza_name][key]
        return stanza

    def _delete_creds(self, stanza):
        """
        :stanza: if there are keys of self._keys and the keys are in stanza,
        delete the encrypted creds
        """

        if self._keys is None:
            return

        for key in self._keys:
            if key in stanza:
                self._cred_mgr.delete(stanza["name"])
                break



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
A timer queue implementation
"""

from future import standard_library

standard_library.install_aliases()
from builtins import object
import threading
import queue
from time import time
import traceback

from splunktalib.timer import Timer
from splunktalib.common import log


class TimerQueue(object):
    """
    A timer queue implementation, runs a separate thread to handle timers
    """

    import sortedcontainers as sc

    def __init__(self):
        self._timers = TimerQueue.sc.SortedSet()
        self._cancelling_timers = {}
        self._lock = threading.Lock()
        self._wakeup_queue = queue.Queue()
        self._thr = threading.Thread(target=self._check_and_execute)
        self._started = False

    def start(self):
        """
        Start the timer queue to make it start function
        """

        if self._started:
            return
        self._started = True

        self._thr.start()
        log.logger.info("TimerQueue started.")

    def tear_down(self):
        if not self._started:
            return
        self._started = True
        self._wakeup(None)
        self._thr.join()

    def add_timer(self, callback, when, interval):
        """
        Add timer to the queue
        """

        timer = Timer(callback, when, interval)
        with self._lock:
            self._timers.add(timer)
        self._wakeup()
        return timer

    def remove_timer(self, timer):
        """
        Remove timer from the queue.
        """

        with self._lock:
            try:
                self._timers.remove(timer)
            except ValueError:
                log.logger.info(
                    "Timer=%s is not in queue, move it to cancelling " "list",
                    timer.ident(),
                )
            else:
                self._cancelling_timers[timer.ident()] = timer

    def _check_and_execute(self):
        wakeup_queue = self._wakeup_queue
        while 1:
            (next_expired_time, expired_timers) = self._get_expired_timers()
            for timer in expired_timers:
                try:
                    timer()
                except Exception:
                    log.logger.error(traceback.format_exc())

            self._reset_timers(expired_timers)

            # Calc sleep time
            if next_expired_time:
                now = time()
                if now < next_expired_time:
                    sleep_time = next_expired_time - now
                else:
                    sleep_time = 0.1
            else:
                sleep_time = 1

            try:
                wakeup = wakeup_queue.get(timeout=sleep_time)
                if wakeup is None:
                    break
            except queue.Empty:
                pass
        log.logger.info("TimerQueue stopped.")

    def _get_expired_timers(self):
        next_expired_time = 0
        now = time()
        expired_timers = []
        with self._lock:
            for timer in self._timers:
                if timer.get_expiration() <= now:
                    expired_timers.append(timer)

            if expired_timers:
                del self._timers[: len(expired_timers)]

            if self._timers:
                next_expired_time = self._timers[0].get_expiration()
        return (next_expired_time, expired_timers)

    def _reset_timers(self, expired_timers):
        has_new_timer = False
        with self._lock:
            cancelling_timers = self._cancelling_timers
            for timer in expired_timers:
                if timer.ident() in cancelling_timers:
                    log.logger.INFO("Timer=%s has been cancelled", timer.ident())
                    continue
                elif timer.get_interval():
                    # Repeated timer
                    timer.update_expiration()
                    self._timers.add(timer)
                    has_new_timer = True
            cancelling_timers.clear()

        if has_new_timer:
            self._wakeup()

    def _wakeup(self, something="not_None"):
        self._wakeup_queue.put(something)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import os
import threading
import time
import traceback

from splunktalib.common import log


class OrphanProcessChecker(object):
    def __init__(self, callback=None):
        """
        Only work for Linux platform. On Windows platform, is_orphan is always
        False
        """

        if os.name == "nt":
            self._ppid = 0
        else:
            self._ppid = os.getppid()
        self._callback = callback

    def is_orphan(self):
        if os.name == "nt":
            return False
        res = self._ppid != os.getppid()
        if res:
            log.logger.warn("Process=%s has become orphan", os.getpid())
        return res

    def check_orphan(self):
        res = self.is_orphan()
        if res and self._callback:
            self._callback()
        return res


class OrphanProcessMonitor(object):
    def __init__(self, callback):
        self._checker = OrphanProcessChecker(callback)
        self._thr = threading.Thread(target=self._do_monitor)
        self._thr.daemon = True
        self._started = False

    def start(self):
        if self._started:
            return
        self._started = True

        self._thr.start()

    def stop(self):
        self._started = False

    def _do_monitor(self):
        while self._started:
            try:
                res = self._checker.check_orphan()
                if res:
                    break
                time.sleep(1)
            except Exception:
                log.logger.error(
                    "Failed to monitor orphan process, reason=%s",
                    traceback.format_exc(),
                )



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from future import standard_library

standard_library.install_aliases()
import os
import os.path as op
import subprocess
from configparser import ConfigParser
from io import StringIO

from splunktalib.common import util as scu


def make_splunkhome_path(parts):
    """
    create a path string by the several parts of the path
    """

    relpath = os.path.normpath(os.path.join(*parts))

    basepath = os.environ["SPLUNK_HOME"]  # Assume SPLUNK_HOME env has been set

    fullpath = os.path.normpath(os.path.join(basepath, relpath))

    # Check that we haven't escaped from intended parent directories.
    if os.path.relpath(fullpath, basepath)[0:2] == "..":
        raise ValueError(
            'Illegal escape from parent directory "%s": %s' % (basepath, fullpath)
        )

    return fullpath


def get_splunk_bin():
    if os.name == "nt":
        splunk_bin = "splunk.exe"
    else:
        splunk_bin = "splunk"
    return make_splunkhome_path(("bin", splunk_bin))


def get_appname_from_path(absolute_path):
    return scu.get_appname_from_path(absolute_path)


def _get_merged_conf_raw(conf_name):
    """
    :conf_name: configure file name
    :return: raw output of all contents for the same conf file
    Note: it depends on SPLUNK_HOME env variable
    """

    assert conf_name

    if conf_name.endswith(".conf"):
        conf_name = conf_name[:-5]

    # FIXME dynamically caculate SPLUNK_HOME
    btool_cli = [op.join(os.environ["SPLUNK_HOME"], "bin", "btool"), conf_name, "list"]

    try:
        p = subprocess.Popen(btool_cli, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
    except OSError:
        raise

    return out


def _get_conf_stanzas(conf_name):
    """
    :return: {stanza_name: stanza_configs}, dict of dict
    """

    res = _get_merged_conf_raw(conf_name)
    res = StringIO(res)
    parser = ConfigParser()
    parser.optionxform = str
    parser.readfp(res)
    res = {}
    for section in parser.sections():
        res[section] = {item[0]: item[1] for item in parser.items(section)}
    return res


def get_splunkd_uri():
    if "SPLUNKD_URI" in os.environ:
        return os.environ["SPLUNKD_URI"]
    else:
        server_conf = _get_conf_stanzas("server")
        if server_conf["sslConfig"]["enableSplunkdSSL"].lower() == "true":
            http = "https://"
        else:
            http = "http://"

        web_conf = _get_conf_stanzas("web")
        host_port = web_conf["settings"]["mgmtHostPort"]
        splunkd_uri = "{}{}".format(http, host_port)

        if os.environ.get("SPLUNK_BINDIP"):
            bip = os.environ["SPLUNK_BINDIP"]
            port_idx = bip.rfind(":")
            if port_idx > 0:
                bip = bip[:port_idx]
            port = host_port[host_port.rfind(":") :]
            splunkd_uri = "{}{}{}".format(http, bip, port)
        return splunkd_uri



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
This module hanles configuration related stuff
"""

from builtins import object
import os.path as op

import splunktalib.conf_manager.conf_endpoints as scmc
import splunktalib.conf_manager.data_input_endpoints as scmdi
import splunktalib.conf_manager.property_endpoints as scmp
import splunktalib.conf_manager.request as req


def conf_file2name(conf_file):
    conf_name = op.basename(conf_file)
    if conf_name.endswith(".conf"):
        conf_name = conf_name[:-5]
    return conf_name


class ConfManager(object):
    def __init__(self, splunkd_uri, session_key, owner="nobody", app_name="-"):
        """
        :app_name: when creating conf stanza, app_name is required to set not
        to "-"
        :owner: when creating conf stanza, app_name is required to set not
        to "-"
        """

        self.splunkd_uri = splunkd_uri
        self.session_key = session_key
        self.owner = owner
        self.app_name = app_name

    def set_appname(self, appname):
        """
        This are cases we need edit/remove/create confs in different app
        context. call this interface to switch app context before manipulate
        the confs in different app context
        """

        self.app_name = appname

    def all_stanzas(self, conf_name, do_reload=False, ret_metadata=False):
        """
        :return: a list of dict stanza objects if successful.
                 Otherwise raise exception
        """

        if do_reload:
            self.reload_conf(conf_name)

        stanzas = scmc.get_conf(self.splunkd_uri, self.session_key, "-", "-", conf_name)
        return self._delete_metadata(stanzas, ret_metadata)

    def all_stanzas_as_dicts(self, conf_name, do_reload=False, ret_metadata=False):
        """
        :return: a dict of dict stanza objects if successful.
                 otherwise raise exception
        """

        stanzas = self.all_stanzas(conf_name, do_reload, ret_metadata)
        return {stanza["name"]: stanza for stanza in stanzas}

    def get_stanza(self, conf_name, stanza, do_reload=False, ret_metadata=False):
        """
        @return dict if success otherwise raise exception
        """

        if do_reload:
            self.reload_conf(conf_name)

        stanzas = scmc.get_conf(
            self.splunkd_uri, self.session_key, "-", "-", conf_name, stanza
        )
        stanzas = self._delete_metadata(stanzas, ret_metadata)
        return stanzas[0]

    def reload_conf(self, conf_name):
        scmc.reload_conf(self.splunkd_uri, self.session_key, "-", conf_name)

    def enable_conf(self, conf_name, stanza):
        scmc.operate_conf(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            conf_name,
            stanza,
            "enable",
        )

    def disable_conf(self, conf_name, stanza):
        scmc.operate_conf(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            conf_name,
            stanza,
            "disable",
        )

    def get_property(self, conf_name, stanza, key, do_reload=False):
        if do_reload:
            self.reload_conf(conf_name)

        return scmp.get_property(
            self.splunkd_uri, self.session_key, "-", "-", conf_name, stanza, key
        )

    def stanza_exist(self, conf_name, stanza):
        return scmc.stanza_exist(
            self.splunkd_uri, self.session_key, "-", "-", conf_name, stanza
        )

    def create_stanza(self, conf_name, stanza, key_values):
        scmc.create_stanza(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            conf_name,
            stanza,
            key_values,
        )

    def update_stanza(self, conf_name, stanza, key_values):
        scmc.update_stanza(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            conf_name,
            stanza,
            key_values,
        )

    def delete_stanza(self, conf_name, stanza):
        scmc.delete_stanza(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            conf_name,
            stanza,
        )

    def create_properties(self, conf_name, stanza):
        scmp.create_properties(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            conf_name,
            stanza,
        )

    def update_properties(self, conf_name, stanza, key_values):
        scmp.update_properties(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            conf_name,
            stanza,
            key_values,
        )

    def delete_stanzas(self, conf_name, stanzas):
        """
        :param stanzas: list of stanzas
        :return: list of failed stanzas
        """

        failed_stanzas = []
        for stanza in stanzas:
            try:
                self.delete_stanza(conf_name, stanza)
            except Exception:
                failed_stanzas.append(stanza)
        return failed_stanzas

    # data input management
    def create_data_input(self, input_type, name, key_values=None):
        scmdi.create_data_input(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            input_type,
            name,
            key_values,
        )

    def update_data_input(self, input_type, name, key_values):
        scmdi.update_data_input(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            input_type,
            name,
            key_values,
        )

    def delete_data_input(self, input_type, name):
        scmdi.delete_data_input(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            input_type,
            name,
        )

    def get_data_input(self, input_type, name=None, do_reload=False):
        if do_reload:
            self.reload_data_input(input_type)

        return scmdi.get_data_input(
            self.splunkd_uri, self.session_key, "-", "-", input_type, name
        )

    def reload_data_input(self, input_type):
        scmdi.reload_data_input(
            self.splunkd_uri, self.session_key, "-", "-", input_type
        )

    def enable_data_input(self, input_type, name):
        scmdi.operate_data_input(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            input_type,
            name,
            "enable",
        )

    def disable_data_input(self, input_type, name):
        scmdi.operate_data_input(
            self.splunkd_uri,
            self.session_key,
            self.owner,
            self.app_name,
            input_type,
            name,
            "disable",
        )

    def data_input_exist(self, input_type, name):
        try:
            result = self.get_data_input(input_type, name)
        except req.ConfNotExistsException:
            return False

        return result is not None

    def all_data_input_stanzas(self, input_type, do_reload=False, ret_metadata=False):
        stanzas = self.get_data_input(input_type, do_reload=do_reload)
        for stanza in stanzas:
            if "eai:acl" in stanza and "app" in stanza["eai:acl"]:
                stanza["appName"] = stanza["eai:acl"]["app"]
                stanza["userName"] = stanza["eai:acl"].get("owner", "nobody")
        return self._delete_metadata(stanzas, ret_metadata)

    def get_data_input_stanza(
        self, input_type, name, do_reload=False, ret_metadata=False
    ):

        stanzas = self.get_data_input(input_type, name, do_reload)
        stanzas = self._delete_metadata(stanzas, ret_metadata)
        return stanzas[0]

    def delete_data_input_stanzas(self, input_type, names):
        """
        :param stanzas: list of stanzas
        :return: list of failed stanzas
        """

        failed_names = []
        for name in names:
            try:
                self.delete_data_input(input_type, name)
            except Exception:
                failed_names.append(name)
        return failed_names

    def _delete_metadata(self, stanzas, ret_metadata):
        if stanzas and not ret_metadata:
            for stanza in stanzas:
                for key in list(stanza.keys()):
                    if key.startswith("eai:"):
                        del stanza[key]
        return stanzas



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import splunktalib.common.xml_dom_parser as xdp
import splunktalib.conf_manager.request as req


class KnowledgeObjectManager(object):
    def __init__(self, splunkd_uri, session_key):
        self.splunkd_uri = splunkd_uri
        self.session_key = session_key

    def apps(self):
        """
        @return: a list of dict containing apps if successfuly otherwise
        otherwise raise exceptions
        """

        uri = "{}/services/apps/local?count=0&offset=0".format(self.splunkd_uri)
        apps = self._do_request(uri, "GET", None, "Failed to get apps")
        return apps

    def indexes(self):
        """
        @return: a list of dict containing indexes if successfuly
        otherwise raise exceptions
        """

        uri = "{}/services/data/indexes/?count=0&offset=0".format(self.splunkd_uri)
        indexes = self._do_request(uri, "GET", None, "Failed to get indexes")
        return indexes

    def _do_request(self, uri, method, payload, err_msg):
        _, content = req.content_request(
            uri, self.session_key, method, payload, err_msg
        )
        return xdp.parse_conf_xml_dom(content)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
Copyright (C) 2005-2019 Splunk Inc. All Rights Reserved.

Commonly used design partten for python user, includes:
  - singleton (Decorator function used to build singleton)
"""
from __future__ import print_function

from functools import wraps


def singleton(class_):
    """
    Singleton decoorator function.
    """
    instances = {}

    @wraps(class_)
    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


class Singleton(type):
    """
    Singleton meta class
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
            print(cls)
        return cls._instances[cls]



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
A wrapper of multiprocessing.pool
"""

from builtins import object
import multiprocessing
from splunktalib.common import log


class ProcessPool(object):
    """
    A simple wrapper of multiprocessing.pool
    """

    def __init__(self, size=0, maxtasksperchild=10000):
        if size <= 0:
            size = multiprocessing.cpu_count()
        self.size = size
        self._pool = multiprocessing.Pool(
            processes=size, maxtasksperchild=maxtasksperchild
        )
        self._stopped = False

    def tear_down(self):
        """
        Tear down the pool
        """

        if self._stopped:
            log.logger.info("ProcessPool has already stopped.")
            return
        self._stopped = True

        self._pool.close()
        self._pool.join()
        log.logger.info("ProcessPool stopped.")

    def apply(self, func, args=(), kwargs={}):
        """
        :param func: callable
        :param args: free params
        :param kwargs: named params
        :return whatever the func returns
        """

        if self._stopped:
            log.logger.info("ProcessPool has already stopped.")
            return None

        return self._pool.apply(func, args, kwargs)

    def apply_async(self, func, args=(), kwargs={}, callback=None):
        """
        :param func: callable
        :param args: free params
        :param kwargs: named params
        :callback: when func is done without exception, call this callack
        :return whatever the func returns
        """

        if self._stopped:
            log.logger.info("ProcessPool has already stopped.")
            return None

        return self._pool.apply_async(func, args, kwargs, callback)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import threading


class Timer(object):
    """
    Timer wraps the callback and timestamp related stuff
    """

    _ident = 0
    _lock = threading.Lock()

    def __init__(self, callback, when, interval, ident=None):
        self._callback = callback
        self._when = when
        self._interval = interval

        if ident is not None:
            self._id = ident
        else:
            with Timer._lock:
                self._id = Timer._ident + 1
                Timer._ident = Timer._ident + 1

    def get_interval(self):
        return self._interval

    def set_interval(self, interval):
        self._interval = interval

    def get_expiration(self):
        return self._when

    def set_initial_due_time(self, when):
        self._when = when

    def update_expiration(self):
        self._when += self._interval

    def __cmp__(self, other):
        if other is None:
            return 1

        self_k = (self.get_expiration(), self.ident())
        other_k = (other.get_expiration(), other.ident())

        if self_k == other_k:
            return 0
        elif self_k < other_k:
            return -1
        else:
            return 1

    def __eq__(self, other):
        return isinstance(other, Timer) and (self.ident() == other.ident())

    def __lt__(self, other):
        return self.__cmp__(other) == -1

    def __gt__(self, other):
        return self.__cmp__(other) == 1

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        return self.__lt__(other) or self.__eq__(other)

    def __ge__(self, other):
        return self.__gt__(other) or self.__eq__(other)

    def __hash__(self):
        return self.ident()

    def __call__(self):
        self._callback()

    def ident(self):
        return self._id



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
Concurrent executor provides concurrent executing function either in
a thread pool or a process pool
"""

from builtins import object
import splunktalib.concurrent.thread_pool as tp
import splunktalib.concurrent.process_pool as pp


class ConcurrentExecutor(object):
    def __init__(self, config):
        """
        :param config: dict like object, contains thread_min_size (int),
                       thread_max_size (int), daemonize_thread (bool),
                       process_size (int)
        """

        self._io_executor = tp.ThreadPool(
            config.get("thread_min_size", 0),
            config.get("thread_max_size", 0),
            config.get("task_queue_size", 1024),
            config.get("daemonize_thread", True),
        )
        self._compute_executor = None
        if config.get("process_size", 0):
            self._compute_executor = pp.ProcessPool(config.get("process_size", 0))

    def start(self):
        self._io_executor.start()

    def tear_down(self):
        self._io_executor.tear_down()
        if self._compute_executor is not None:
            self._compute_executor.tear_down()

    def run_io_func_sync(self, func, args=(), kwargs=None):
        """
        :param func: callable
        :param args: free params
        :param kwargs: named params
        :return whatever the func returns
        """

        return self._io_executor.apply(func, args, kwargs)

    def run_io_func_async(self, func, args=(), kwargs=None, callback=None):
        """
        :param func: callable
        :param args: free params
        :param kwargs: named params
        :calllback: when func is done and without exception, call the callback
        :return whatever the func returns
        """

        return self._io_executor.apply_async(func, args, kwargs, callback)

    def enqueue_io_funcs(self, funcs, block=True):
        """
        run jobs in a fire and forget way, no result will be handled
        over to clients
        :param funcs: tuple/list-like or generator like object, func shall be
                      callable
        """

        return self._io_executor.enqueue_funcs(funcs, block)

    def run_compute_func_sync(self, func, args=(), kwargs={}):
        """
        :param func: callable
        :param args: free params
        :param kwargs: named params
        :return whatever the func returns
        """

        assert self._compute_executor is not None
        return self._compute_executor.apply(func, args, kwargs)

    def run_compute_func_async(self, func, args=(), kwargs={}, callback=None):
        """
        :param func: callable
        :param args: free params
        :param kwargs: named params
        :calllback: when func is done and without exception, call the callback
        :return whatever the func returns
        """

        assert self._compute_executor is not None
        return self._compute_executor.apply_async(func, args, kwargs, callback)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import str
import splunktalib.common.xml_dom_parser as xdp
import splunktalib.common.util as util
from splunktalib.conf_manager.request import content_request


INPUT_ENDPOINT = "%s/servicesNS/%s/%s/data/inputs/%s"


def _input_endpoint_ns(uri, owner, app, input_type):
    return INPUT_ENDPOINT % (uri, owner, app, input_type)


def reload_data_input(
    splunkd_uri, session_key, owner, app_name, input_type, throw=False
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param input_type: name of the input type.
                       if it is a script input, the input is "script",
                       for modinput, say snow, the input is "snow"
    """

    uri = _input_endpoint_ns(splunkd_uri, owner, app_name, input_type)
    uri += "/_reload"
    msg = "Failed to reload data input in app=%s: %s" % (app_name, input_type)
    try:
        content_request(uri, session_key, "GET", None, msg)
    except Exception:
        if throw:
            raise


def create_data_input(
    splunkd_uri, session_key, owner, app_name, input_type, name, key_values
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param input_type: name of the input type.
                       if it is a script input, the input is "script",
                       for modinput, say snow, the input is "snow"
    :param name: The name of the input stanza to create.
                 i.e. stanza [<input_type>://<name>] will be created.
    :param key_values: a K-V dict of details in the data input stanza.
    :return: None on success else raise exception
    """

    key_values["name"] = str(name).encode("utf-8")
    uri = _input_endpoint_ns(splunkd_uri, owner, app_name, input_type)
    msg = "Failed to create data input in app=%s: %s://%s" % (
        app_name,
        input_type,
        name,
    )
    content_request(uri, session_key, "POST", key_values, msg)


def get_data_input(splunkd_uri, session_key, owner, app_name, input_type, name=None):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param input_type: name of the input type.
                       if it is a script input, the input is "script",
                       for modinput, say snow, the input is "snow"
    :param name: The name of the input stanza to create.
                 i.e. stanza [<input_type>://<name>] will be deleted.
    :return: a list of stanzas in the input type, including metadata
    """

    uri = _input_endpoint_ns(splunkd_uri, owner, app_name, input_type)
    if name:
        uri += "/" + util.format_stanza_name(name)

    # get all the stanzas at one time
    uri += "?count=0&offset=0"

    msg = "Failed to get data input in app=%s: %s://%s" % (
        app_name,
        input_type,
        name if name else name,
    )
    content = content_request(uri, session_key, "GET", None, msg)
    return xdp.parse_conf_xml_dom(content)


def update_data_input(
    splunkd_uri, session_key, owner, app_name, input_type, name, key_values
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param input_type: name of the input type.
                       if it is a script input, the input is "script",
                       for modinput, say snow, the input is "snow"
    :param name: The name of the input stanza to create.
                 i.e. stanza [<input_type>://<name>] will be updated.
    :param key_values: a K-V dict of details in the data input stanza.
    :return: raise exception when failure
    """

    if "name" in key_values:
        del key_values["name"]
    uri = _input_endpoint_ns(splunkd_uri, owner, app_name, input_type)
    uri += "/" + util.format_stanza_name(name)
    msg = "Failed to update data input in app=%s: %s://%s" % (
        app_name,
        input_type,
        name,
    )
    content_request(uri, session_key, "POST", key_values, msg)


def delete_data_input(splunkd_uri, session_key, owner, app_name, input_type, name):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param input_type: name of the input type.
                       if it is a script input, the input is "script",
                       for modinput, say snow, the input is "snow"
    :param name: The name of the input stanza to create.
                 i.e. stanza [<input_type>://<name>] will be deleted.
    :return raise exception when failed
    """

    uri = _input_endpoint_ns(splunkd_uri, owner, app_name, input_type)
    uri += "/" + util.format_stanza_name(name)
    msg = "Failed to delete data input in app=%s: %s://%s" % (
        app_name,
        input_type,
        name,
    )
    content_request(uri, session_key, "DELETE", None, msg)


def operate_data_input(
    splunkd_uri, session_key, owner, app_name, input_type, name, operation
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param input_type: name of the input type.
                       if it is a script input, the input is "script",
                       for modinput, say snow, the input is "snow"
    :param name: The name of the input stanza to create.
                 i.e. stanza [<input_type>://<name>] will be operated.
    :param operation: must be "disable" or "enable"
    """

    assert operation in ("disable", "enable")
    uri = _input_endpoint_ns(splunkd_uri, owner, app_name, input_type)
    uri += "/%s/%s" % (util.format_stanza_name(name), operation)
    msg = "Failed to %s data input in app=%s: %s://%s" % (
        operation,
        app_name,
        input_type,
        name,
    )
    content_request(uri, session_key, "POST", None, msg)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

# import unittest
# import subprocess
# import json
# import pytest

# @pytest.mark.usefixtures("splunk")
# @pytest.mark.usefixtures("splunk_setup")
# class TestCreateFunction(unittest.TestCase):
#     def setUp(self):
#         """
#         Account entity definition

#         "entity": [
#             {
#                 "field": "name",
#                 "label": "Name",
#                 "type": "text",
#                 "required": true,
#                 "help": "Enter a unique name for each Crowdstrike falcon host account.",
#                 "validators": [
#                     {
#                         "type": "string",
#                         "minLength": 1,
#                         "maxLength": 50,
#                         "errorMsg": "Length of Name is restricted from 1 to 50."
#                     },
#                     {
#                         "type": "regex",
#                         "pattern": "^\\w+$",
#                         "errorMsg": "Characters of Name should match regex ^\\w+$ ."
#                     }
#                 ]
#             },
#             {
#                 "field": "endpoint",
#                 "label": "Endpoint",
#                 "type": "text",
#                 "required": true,
#                 "defaultValue": "https://firehose.crowdstrike.com/sensors/entities/datafeed/v1",
#                 "options": {
#                     "enabled": false,
#                     "placeholder": "https://firehose.crowdstrike.com/sensors/entities/datafeed/v1"
#                 }
#             },
#             {
#                 "field": "api_uuid",
#                 "label": "API UUID",
#                 "type": "text",
#                 "required": false,
#                 "defaultValue": "123456789"
#             },
#             {
#                 "field": "api_uuid2",
#                 "label": "API UUID2",
#                 "type": "text",
#                 "required": false,
#                 "defaultValue": "123456789"
#             },
#             {
#                 "field": "test_optional_password",
#                 "label": "test_optional_password",
#                 "type": "text",
#                 "required": false,
#                 "encrypted": true
#             }
#         ]
#         """
#         self.endpoint = "https://localhost:8089/servicesNS/nobody/Splunk_TA_UCCExample/splunk_ta_uccexample_account"
#         self.conf_endpoint = (
#             "https://localhost:8089/servicesNS/nobody/Splunk_TA_UCCExample/"
#             "configs/conf-splunk_ta_uccexample_account"
#         )
#         self.password_endpoint = (
#             "https://localhost:8089/servicesNS/nobody/-/storage/passwords"
#         )
#         self.test_item_name = "1"
#         self.PASSWORD = "******"

#     def create_item(self, data, via_conf=False, name=None):
#         if via_conf:
#             endpoint = self.conf_endpoint
#         else:
#             endpoint = self.endpoint

#         if not name:
#             name = self.test_item_name
#         create_cmd = " ".join([self.get_prefix("POST"), endpoint, data])
#         output = subprocess.check_output(create_cmd.split(" "))
#         # check if it's created successfully
#         self.assertTrue("<title>{name}</title>".format(name=name) in str(output))

#     def update_item(self, data, via_conf=False, name=None):
#         if via_conf:
#             endpoint = self.conf_endpoint
#         else:
#             endpoint = self.endpoint

#         if not name:
#             name = self.test_item_name
#         update_cmd = " ".join([self.get_prefix("POST"), endpoint + "/" + name, data])
#         output = subprocess.check_output(update_cmd.split(" "))
#         # check the update is done
#         self.assertTrue("<title>{name}</title>".format(name=name) in output)

#     def delete_item(self, name=None):
#         if not name:
#             name = self.test_item_name
#         delete_cmd = " ".join(
#             [
#                 self.get_prefix("DELETE"),
#                 self.endpoint + "/" + name,
#                 "-d output_mode=json",
#             ]
#         )
#         output = subprocess.check_output(delete_cmd.split(" "))
#         try:
#             response = json.loads(output)
#             entry = response["entry"]
#         except ValueError:
#             self.fail("Fail to get the entry")
#         if not name:
#             self.assertFalse(entry)

#     def get_prefix(self, method):
#         return "curl -k -u admin:admin -X {method}".format(method=method)

#     def get_item_content(self, clear_password=False):
#         get_cmd = " ".join(
#             [
#                 self.get_prefix("GET"),
#                 self.endpoint + "/" + self.test_item_name,
#                 "-d output_mode=json",
#             ]
#         )
#         if clear_password:
#             get_cmd = " ".join([get_cmd, "-d --cred--=1"])
#         response = subprocess.check_output(get_cmd.split(" "))
#         try:
#             item = json.loads(response)
#         except ValueError:
#             self.fail("Could not get the item %s" % self.test_item_name)

#         # get the item content
#         try:
#             content = item["entry"][0]["content"]
#         except KeyError:
#             self.fail("Could not get the item content")
#         return content

#     def testUpdateWithStars(self):
#         self.create_item(
#             data="-d name={name} -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
#                 name=self.test_item_name
#             )
#         )
#         # update with empty password
#         self.update_item(
#             data="-d endpoint=1 -d endpoint=1 -d api_key=1 -d test_optional_password="
#         )
#         content = self.get_item_content(clear_password=True)
#         self.assertEqual(content["test_optional_password"], "")
#         # update with magic password '******'
#         self.update_item(
#             data="-d endpoint=1 -d endpoint=1 -d api_key=1 -d test_optional_password=*******"
#         )
#         content = self.get_item_content(clear_password=True)
#         self.assertEqual(content["test_optional_password"], "*******")
#         self.delete_item()


# if __name__ == "__main__":
#     unittest.main()



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

import pytest
import json
import os
from splunk_add_on_ucc_framework.splunktaucclib.global_config import (
    GlobalConfig,
    GlobalConfigSchema,
)


def test_SaveConfig(get_session_key):

    payload = {
        "account": [
            {"endpoint": "abc", "api_key": "1", "name": "account0"},
            {"endpoint": "abc", "api_key": "1", "name": "account1"},
        ],
        "settings": [
            {"loglevel": "INFO", "name": "logging"},
            {"proxy_type": "http", "proxy_password": "", "name": "proxy"},
        ],
    }
    basedir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    with open(
        os.path.join(os.path.dirname(basedir), "splunktaucclib/globalConfig.json")
    ) as f:
        json_schema = "".join([l for l in f])

    schema = GlobalConfigSchema(json.loads(json_schema))

    sk = get_session_key[1]

    global_config = GlobalConfig(get_session_key[0], sk, schema)
    save_results = global_config.save(payload)
    for result in save_results:
        assert result is None



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import range
from builtins import object
import os.path as op
import os
import json
import time
import splunktalib.kv_client as kvc
from splunktalib.common import util
from splunktalib.common import log
import threading


def get_state_store(
    meta_configs,
    appname,
    collection_name="talib_states",
    use_kv_store=False,
    use_cached_store=False,
):
    if util.is_true(use_kv_store):
        return StateStore(meta_configs, appname, collection_name)
    elif util.is_true(use_cached_store):
        return CachedFileStateStore(meta_configs, appname)
    else:
        return FileStateStore(meta_configs, appname)


class BaseStateStore(object):
    def __init__(self, meta_configs, appname):
        self._meta_configs = meta_configs
        self._appname = appname

    def update_state(self, key, states):
        pass

    def get_state(self, key):
        pass

    def delete_state(self, key):
        pass

    def close(self, key=None):
        pass


class StateStore(BaseStateStore):
    def __init__(self, meta_configs, appname, collection_name="talib_states"):
        """
        :meta_configs: dict like and contains checkpoint_dir, session_key,
         server_uri etc
        :app_name: the name of the app
        :collection_name: the collection name to be used.
        Don"t use other method to visit the collection if you are using
         StateStore to visit it.
        """
        super(StateStore, self).__init__(meta_configs, appname)

        # State cache is a dict from _key to value
        self._states_cache = {}
        self._kv_client = None
        self._collection = collection_name
        self._kv_client = kvc.KVClient(
            meta_configs["server_uri"], meta_configs["session_key"]
        )
        kvc.create_collection(self._kv_client, self._collection, self._appname)
        self._load_states_cache()

    def update_state(self, key, states):
        """
        :state: Any JSON serializable
        :return: None if successful, otherwise throws exception
        """

        if key not in self._states_cache:
            self._kv_client.insert_collection_data(
                self._collection,
                {"_key": key, "value": json.dumps(states)},
                self._appname,
            )
        else:
            self._kv_client.update_collection_data(
                self._collection, key, {"value": json.dumps(states)}, self._appname
            )
        self._states_cache[key] = states

    def get_state(self, key=None):
        if key:
            return self._states_cache.get(key, None)
        return self._states_cache

    def delete_state(self, key=None):
        if key:
            self._delete_state(key)
        else:
            [self._delete_state(_key) for _key in list(self._states_cache.keys())]

    def _delete_state(self, key):
        if key not in self._states_cache:
            return

        self._kv_client.delete_collection_data(self._collection, key, self._appname)
        del self._states_cache[key]

    def _load_states_cache(self):
        states = self._kv_client.get_collection_data(
            self._collection, None, self._appname
        )
        if not states:
            return

        for state in states:
            if "value" in state:
                value = state["value"]
            else:
                value = state

            try:
                value = json.loads(value)
            except Exception:
                pass

            self._states_cache[state["_key"]] = value


class FileStateStore(BaseStateStore):
    def __init__(self, meta_configs, appname):
        """
        :meta_configs: dict like and contains checkpoint_dir, session_key,
        server_uri etc
        """

        super(FileStateStore, self).__init__(meta_configs, appname)

    def update_state(self, key, states):
        """
        :state: Any JSON serializable
        :return: None if successful, otherwise throws exception
        """

        fname = op.join(self._meta_configs["checkpoint_dir"], key)
        with open(fname + ".new", "w") as jsonfile:
            json.dump(states, jsonfile)

        if op.exists(fname):
            os.remove(fname)

        os.rename(fname + ".new", fname)
        # commented this to disable state cache for local file
        # if key not in self._states_cache:
        # self._states_cache[key] = {}
        # self._states_cache[key] = states

    def get_state(self, key):
        fname = op.join(self._meta_configs["checkpoint_dir"], key)
        if op.exists(fname) and op.isfile(fname):
            with open(fname) as jsonfile:
                state = json.load(jsonfile)
                # commented this to disable state cache for local file
                # self._states_cache[key] = state
                return state
        else:
            return None

    def delete_state(self, key):
        fname = op.join(self._meta_configs["checkpoint_dir"], key)
        if op.exists(fname):
            os.remove(fname)


class CachedFileStateStore(BaseStateStore):
    def __init__(self, meta_configs, appname, max_cache_seconds=5):
        """
        :meta_configs: dict like and contains checkpoint_dir, session_key,
        server_uri etc
        """

        super(CachedFileStateStore, self).__init__(meta_configs, appname)
        self._states_cache = {}  # item: time, dict
        self._states_cache_lmd = {}  # item: time, dict
        self.max_cache_seconds = max_cache_seconds
        self._lock = threading.RLock()

    def update_state(self, key, states):
        with self._lock:
            now = time.time()
            if key in self._states_cache:
                last = self._states_cache_lmd[key][0]
                if now - last >= self.max_cache_seconds:
                    self.update_state_flush(now, key, states)
            else:
                self.update_state_flush(now, key, states)
            self._states_cache[key] = (now, states)

    def update_state_flush(self, now, key, states):
        """
        :state: Any JSON serializable
        :return: None if successful, otherwise throws exception
        """
        for i in range(3):
            try:
                with self._lock:
                    self._states_cache_lmd[key] = (now, states)
                    fname = op.join(self._meta_configs["checkpoint_dir"], key)
                    with open(fname + ".new", "w") as jsonfile:
                        json.dump(states, jsonfile)

                    if op.exists(fname):
                        os.remove(fname)

                    os.rename(fname + ".new", fname)
            except Exception:
                log.logger.exception("Failed to update checkpoint:")
                time.sleep(1)
                continue
            else:
                return
        raise Exception("Failed to update checkpoint")

    def get_state(self, key):
        with self._lock:
            if key in self._states_cache:
                return self._states_cache[key][1]

            fname = op.join(self._meta_configs["checkpoint_dir"], key)
            if op.exists(fname):
                with open(fname) as jsonfile:
                    state = json.load(jsonfile)
                    now = time.time()
                    self._states_cache[key] = now, state
                    self._states_cache_lmd[key] = now, state
                    return state
            else:
                return None

    def delete_state(self, key):
        with self._lock:
            fname = op.join(self._meta_configs["checkpoint_dir"], key)
            if op.exists(fname):
                os.remove(fname)

            if self._states_cache.get(key):
                del self._states_cache[key]
            if self._states_cache_lmd.get(key):
                del self._states_cache_lmd[key]

    def close(self, key=None):
        for i in range(3):
            try:
                with self._lock:
                    if not key:
                        for k, (t, s) in self._states_cache.items():
                            self.update_state_flush(t, k, s)
                        self._states_cache.clear()
                        self._states_cache_lmd.clear()
                    elif key in self._states_cache:
                        self.update_state_flush(
                            self._states_cache[key][0], key, self._states_cache[key][1]
                        )
                        del self._states_cache[key]
                        del self._states_cache_lmd[key]
            except Exception:
                log.logger.exception("Failed to close checkpoint:")
                time.sleep(1)
                continue
            else:
                return



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from splunktalib.conf_manager.request import content_request
import splunktalib.common.util as util

PROPERTY_ENDPOINT = "%s/servicesNS/%s/%s/properties/%s"


def _property_endpoint_ns(uri, owner, app, conf_name):
    return PROPERTY_ENDPOINT % (uri, owner, app, conf_name)


def create_properties(splunkd_uri, session_key, owner, app_name, conf_name, stanza):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param conf_name: the name of the conf file, e.g. "props"
    :param stanza: stanza name, e.g. "aws:cloudtrail"
    :return: None on success else raise exception
    """

    uri = _property_endpoint_ns(splunkd_uri, owner, app_name, conf_name)
    msg = "Properties: failed to create stanza=%s in conf=%s" % (stanza, conf_name)
    payload = {"__stanza": stanza}
    content_request(uri, session_key, "POST", payload, msg)


def get_property(splunkd_uri, session_key, owner, app_name, conf_name, stanza, key):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param conf_name: the name of the conf file, e.g. "props"
    :param stanza: stanza name, e.g. "aws:cloudtrail"
    :param key: the property name
    :return: the property value
    """

    uri = _property_endpoint_ns(splunkd_uri, owner, app_name, conf_name)
    uri += "/%s/%s" % (util.format_stanza_name(stanza), key)
    msg = "Properties: failed to get conf=%s, stanza=%s, key=%s" % (
        conf_name,
        stanza,
        key,
    )
    return content_request(uri, session_key, "GET", None, msg)


def update_properties(
    splunkd_uri, session_key, owner, app_name, conf_name, stanza, key_values
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param conf_name: the name of the conf file, e.g. "props"
    :param stanza: stanza name, e.g. "aws:cloudtrail"
    :param key_values: the key-value dict of the stanza
    :return: raise exception when failed
    """

    uri = _property_endpoint_ns(splunkd_uri, owner, app_name, conf_name)
    uri += "/" + util.format_stanza_name(stanza)
    msg = "Properties: failed to update conf=%s, stanza=%s" % (conf_name, stanza)

    has_name = False
    if "name" in key_values:
        has_name = True
        name = key_values["name"]
        del key_values["name"]

    content_request(uri, session_key, "POST", key_values, msg)

    if has_name:
        key_values["name"] = name



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import object
import splunktalib.rest as rest
import splunktalib.common.xml_dom_parser as xdp


def _do_rest(uri, session_key):
    resp, content = rest.splunkd_request(uri, session_key)
    if resp is None:
        return None

    if resp.status not in (200, 201):
        return None

    stanza_objs = xdp.parse_conf_xml_dom(content)
    if not stanza_objs:
        return None

    return stanza_objs[0]


class ServerInfo(object):
    def __init__(self, splunkd_uri, session_key):
        uri = "{}/services/server/info".format(splunkd_uri)
        server_info = _do_rest(uri, session_key)
        if server_info is None:
            raise Exception("Failed to init ServerInfo")

        self._server_info = server_info

    def is_captain(self):
        """
        :return: True if splunkd_uri is captain otherwise False
        """

        return "shc_captain" in self._server_info["server_roles"]

    def is_search_head(self):
        for sh in ("search_head", "cluster_search_head"):
            if sh in self._server_info["server_roles"]:
                return True
        return False

    def is_shc_member(self):
        for sh in ("shc_member", "shc_captain"):
            if sh in self._server_info["server_roles"]:
                return True
        return False

    def version(self):
        return self._server_info["version"]

    def to_dict(self):
        return self._server_info


if __name__ == "__main__":
    import splunktalib.credentials as cred

    sp_uri = "https://localhost:8089"
    skey = cred.CredentialManager.get_session_key("admin", "admin")
    si = ServerInfo(sp_uri, skey)
    assert not si.is_captain()

    sp_uri = "https://qa-systest-03.sv.splunk.com:1901"
    skey = cred.CredentialManager.get_session_key("admin", "notchagneme")
    si = ServerInfo(sp_uri, skey)
    assert si.is_captain()
    assert si.is_search_head()

    sp_uri = "https://qa-systest-01.sv.splunk.com:1901"
    skey = cred.CredentialManager.get_session_key("admin", "notchagneme")
    si = ServerInfo(sp_uri, skey)
    assert not si.is_captain()
    assert si.is_search_head()



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import str
import splunktalib.common.xml_dom_parser as xdp
import splunktalib.common.util as util
from splunktalib.conf_manager.request import content_request

CONF_ENDPOINT = "%s/servicesNS/%s/%s/configs/conf-%s"


def _conf_endpoint_ns(uri, owner, app, conf_name):
    return CONF_ENDPOINT % (uri, owner, app, conf_name)


def reload_conf(splunkd_uri, session_key, app_name, conf_name, throw=False):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param conf_names: a list of the name of the conf file, e.g. ["props"]
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    """

    uri = _conf_endpoint_ns(splunkd_uri, "nobody", app_name, conf_name)
    uri += "/_reload"
    msg = "Failed to reload conf in app=%s: %s" % (app_name, conf_name)

    try:
        content_request(uri, session_key, "GET", None, msg)
    except Exception:
        if throw:
            raise


def create_stanza(
    splunkd_uri, session_key, owner, app_name, conf_name, stanza, key_values
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param conf_name: the name of the conf file, e.g. "props"
    :param stanza: stanza name, e.g. "aws:cloudtrail"
    :param key_values: the key-value dict of the stanza
    :return: None on success otherwise throw exception
    """

    uri = _conf_endpoint_ns(splunkd_uri, owner, app_name, conf_name)
    msg = "Failed to create stanza=%s in conf=%s" % (stanza, conf_name)
    payload = {"name": str(stanza).encode("utf-8")}
    for key in key_values:
        if key != "name":
            payload[key] = str(key_values[key])

    content_request(uri, session_key, "POST", payload, msg)


def get_conf(splunkd_uri, session_key, owner, app_name, conf_name, stanza=None):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param conf_name: the name of the conf file, e.g. "props"
    :param stanza: stanza name, e.g. "aws:cloudtrail"
    :return: a list of stanzas in the conf file, including metadata
    """

    uri = _conf_endpoint_ns(splunkd_uri, owner, app_name, conf_name)

    if stanza:
        uri += "/" + util.format_stanza_name(stanza)

    # get all the stanzas at one time
    uri += "?count=0&offset=0"

    msg = "Failed to get stanza=%s in conf=%s" % (
        stanza if stanza else stanza,
        conf_name,
    )
    content = content_request(uri, session_key, "GET", None, msg)
    return xdp.parse_conf_xml_dom(content)


def update_stanza(
    splunkd_uri, session_key, owner, app_name, conf_name, stanza, key_values
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param conf_name: the name of the conf file, e.g. "props"
    :param stanza: stanza name, e.g. "aws:cloudtrail"
    :param key_values: the key-value dict of the stanza
    :return: None on success otherwise raise exception
    """

    uri = _conf_endpoint_ns(splunkd_uri, owner, app_name, conf_name)
    uri += "/" + util.format_stanza_name(stanza)
    msg = "Failed to update stanza=%s in conf=%s" % (stanza, conf_name)
    return content_request(uri, session_key, "POST", key_values, msg)


def delete_stanza(
    splunkd_uri, session_key, owner, app_name, conf_name, stanza, throw=False
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param conf_name: the name of the conf file, e.g. "props"
    :param stanza: stanza name, e.g. "aws:cloudtrail"
    :return: None on success otherwise raise exception
    """

    uri = _conf_endpoint_ns(splunkd_uri, owner, app_name, conf_name)
    uri += "/" + util.format_stanza_name(stanza)
    msg = "Failed to delete stanza=%s in conf=%s" % (stanza, conf_name)
    content_request(uri, session_key, "DELETE", None, msg)


def stanza_exist(splunkd_uri, session_key, owner, app_name, conf_name, stanza):
    try:
        res = get_conf(splunkd_uri, session_key, owner, app_name, conf_name, stanza)
        return len(res) > 0
    except Exception:
        return False


def operate_conf(
    splunkd_uri, session_key, owner, app_name, conf_name, stanza, operation
):
    """
    :param splunkd_uri: splunkd uri, e.g. https://127.0.0.1:8089
    :param session_key: splunkd session key
    :param owner: the owner (ACL user), e.g. "-", "nobody"
    :param app_name: the app"s name, e.g. "Splunk_TA_aws"
    :param conf_name: the name of the conf file, e.g. "props"
    :param stanza: stanza name, e.g. "aws:cloudtrail"
    :param operation: must be "disable" or "enable"
    """

    assert operation in ("disable", "enable")
    uri = _conf_endpoint_ns(splunkd_uri, owner, app_name, conf_name)
    uri += "/%s/%s" % (util.format_stanza_name(stanza), operation)
    msg = "Failed to disable/enable stanza=%s in conf=%s" % (stanza, conf_name)
    content_request(uri, session_key, "POST", None, msg)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

import os
import pytest

pytest_plugins = "pytester"
import urllib.parse
import json
from solnlib.splunk_rest_client import SplunkRestClient
from splunklib import binding


def pytest_configure(config):
    config.addinivalue_line("markers", "external: Test search time only")
    config.addinivalue_line("markers", "docker: Test search time only")


@pytest.fixture(scope="session")
def docker_compose_files(request):
    """
    Get an absolute path to the  `docker-compose.yml` file. Override this
    fixture in your tests if you need a custom location.
    Returns:
        string: the path of the `docker-compose.yml` file
    """
    docker_compose_path = os.path.join(
        str(request.config.invocation_dir), "docker-compose.yml"
    )
    # LOGGER.info("docker-compose path: %s", docker_compose_path)

    return [docker_compose_path]


@pytest.fixture(scope="session")
def get_session_key(splunk, request):

    uri = f'https://{splunk["host"]}:{splunk["port"]}/services/auth/login'
    _rest_client = SplunkRestClient(
        None, "-", "nobody", "https", splunk["host"], splunk["port"]
    )
    try:
        response = _rest_client.http.post(
            uri,
            username=splunk["username"],
            password=splunk["password"],
            output_mode="json",
        )
    except binding.HTTPError as e:
        raise

    return uri, json.loads(response.body.read())["sessionKey"]



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from future import standard_library

standard_library.install_aliases()
from builtins import object
import threading
from time import time
import random
import queue
from splunktalib.common import log


class Scheduler(object):
    """
    A simple scheduler which schedules the periodic or once event
    """

    import sortedcontainers as sc

    max_delay_time = 60

    def __init__(self):
        self._jobs = Scheduler.sc.SortedSet()
        self._wakeup_q = queue.Queue()
        self._lock = threading.Lock()
        self._thr = threading.Thread(target=self._do_jobs)
        self._thr.deamon = True
        self._started = False

    def start(self):
        """
        Start the schduler which will start the internal thread for scheduling
        jobs. Please do tear_down when doing cleanup
        """

        if self._started:
            log.logger.info("Scheduler already started.")
            return
        self._started = True

        self._thr.start()

    def tear_down(self):
        """
        Stop the schduler which will stop the internal thread for scheduling
        jobs.
        """

        if not self._started:
            log.logger.info("Scheduler already tear down.")
            return

        self._wakeup_q.put(True)

    def _do_jobs(self):
        while 1:
            (sleep_time, jobs) = self.get_ready_jobs()
            self._do_execution(jobs)
            try:
                done = self._wakeup_q.get(timeout=sleep_time)
            except queue.Empty:
                pass
            else:
                if done:
                    break
        self._started = False
        log.logger.info("Scheduler exited.")

    def get_ready_jobs(self):
        """
        @return: a 2 element tuple. The first element is the next ready
                 duration. The second element is ready jobs list
        """

        now = time()
        ready_jobs = []
        sleep_time = 1

        with self._lock:
            job_set = self._jobs
            total_jobs = len(job_set)
            for job in job_set:
                if job.get_expiration() <= now:
                    ready_jobs.append(job)

            if ready_jobs:
                del job_set[: len(ready_jobs)]

            for job in ready_jobs:
                if job.get_interval() != 0 and not job.stopped():
                    # repeated job, calculate next due time and enqueue
                    job.update_expiration()
                    job_set.add(job)

            if job_set:
                sleep_time = job_set[0].get_expiration() - now
                if sleep_time < 0:
                    log.logger.warn("Scheduler satuation, sleep_time=%s", sleep_time)
                    sleep_time = 0.1

        if ready_jobs:
            log.logger.info(
                "Get %d ready jobs, next duration is %f, "
                "and there are %s jobs scheduling",
                len(ready_jobs),
                sleep_time,
                total_jobs,
            )

        ready_jobs.sort(key=lambda job: job.get("priority", 0), reverse=True)
        return (sleep_time, ready_jobs)

    def add_jobs(self, jobs):
        with self._lock:
            now = time()
            job_set = self._jobs
            for job in jobs:
                delay_time = random.randrange(0, self.max_delay_time)
                job.set_initial_due_time(now + delay_time)
                job_set.add(job)
        self._wakeup()

    def update_jobs(self, jobs):
        with self._lock:
            job_set = self._jobs
            for njob in jobs:
                job_set.discard(njob)
                job_set.add(njob)
        self._wakeup()

    def remove_jobs(self, jobs):
        with self._lock:
            job_set = self._jobs
            for njob in jobs:
                njob.stop()
                job_set.discard(njob)
        self._wakeup()

    def number_of_jobs(self):
        with self._lock:
            return len(self._jobs)

    def disable_randomization(self):
        self.max_delay_time = 1

    def _wakeup(self):
        self._wakeup_q.put(None)

    def _do_execution(self, jobs):
        for job in jobs:
            job()



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

import unittest
import subprocess
import json
import pytest


@pytest.mark.usefixtures("splunk")
@pytest.mark.usefixtures("splunk_setup")
class TestCreateFunction(unittest.TestCase):
    def setUp(self):
        """
        Account entity definition

        "entity": [
            {
                "field": "name",
                "label": "Name",
                "type": "text",
                "required": true,
                "help": "Enter a unique name for each Crowdstrike falcon host account.",
                "validators": [
                    {
                        "type": "string",
                        "minLength": 1,
                        "maxLength": 50,
                        "errorMsg": "Length of Name is restricted from 1 to 50."
                    },
                    {
                        "type": "regex",
                        "pattern": "^\\w+$",
                        "errorMsg": "Characters of Name should match regex ^\\w+$ ."
                    }
                ]
            },
            {
                "field": "endpoint",
                "label": "Endpoint",
                "type": "text",
                "required": true,
                "defaultValue": "https://firehose.crowdstrike.com/sensors/entities/datafeed/v1",
                "options": {
                    "enabled": false,
                    "placeholder": "https://firehose.crowdstrike.com/sensors/entities/datafeed/v1"
                }
            },
            {
                "field": "api_uuid",
                "label": "API UUID",
                "type": "text",
                "required": false,
                "defaultValue": "123456789"
            },
            {
                "field": "api_uuid2",
                "label": "API UUID2",
                "type": "text",
                "required": false,
                "defaultValue": "123456789"
            },
            {
                "field": "api_key",
                "label": "API Key",
                "type": "text",
                "required": true,
                "encrypted": true
            },
            {
                "field": "test_optional_password",
                "label": "test_optional_password",
                "type": "text",
                "required": false,
                "encrypted": true
            },
            {
                "field": "f1",
                "label": "f1",
                "type": "text",
                "required": false,
                "encrypted": true
            },
            {
                "field": "f2",
                "label": "f2",
                "type": "text",
                "required": false,
                "encrypted": true
            },
            {
                "field": "f3",
                "label": "f3",
                "type": "text",
                "required": false,
                "encrypted": true
            },
            {
                "field": "f4",
                "label": "f4",
                "type": "text",
                "required": false
            },
            {
                "field": "f5",
                "label": "f5",
                "type": "text",
                "required": false
            }
        ]        :return:
        """
        self.endpoint = "https://localhost:8089/servicesNS/nobody/Splunk_TA_UCCExample/splunk_ta_uccexample_account"
        self.conf_endpoint = (
            "https://localhost:8089/servicesNS/nobody/Splunk_TA_UCCExample/"
            "configs/conf-splunk_ta_uccexample_account"
        )
        self.password_endpoint = (
            "https://localhost:8089/servicesNS/nobody/-/storage/passwords"
        )
        self.test_item_name = "1"
        self.PASSWORD = "******"

    def create_item(self, data, via_conf=False, name=None):
        if via_conf:
            endpoint = self.conf_endpoint
        else:
            endpoint = self.endpoint

        if not name:
            name = self.test_item_name
        create_cmd = " ".join([self.get_prefix("POST"), endpoint, data])
        output = subprocess.check_output(create_cmd.split(" ")).decode()
        # check if it's created successfully
        self.assertTrue("<title>{name}</title>".format(name=name) in output)

    def update_item(self, data, via_conf=False, name=None):
        if via_conf:
            endpoint = self.conf_endpoint
        else:
            endpoint = self.endpoint

        if not name:
            name = self.test_item_name
        update_cmd = " ".join([self.get_prefix("POST"), endpoint + "/" + name, data])
        output = subprocess.check_output(update_cmd.split(" "))
        # check the update is done
        self.assertTrue("<title>{name}</title>".format(name=name) in output)

    def delete_item(self, name=None):
        if not name:
            name = self.test_item_name
        delete_cmd = " ".join(
            [
                self.get_prefix("DELETE"),
                self.endpoint + "/" + name,
                "-d output_mode=json",
            ]
        )
        output = subprocess.check_output(delete_cmd.split(" "))
        try:
            response = json.loads(output)
            entry = response["entry"]
        except ValueError:
            self.fail("Fail to get the entry")
        if not name:
            self.assertFalse(entry)

    def get_prefix(self, method):
        return "curl -k -u admin:admin -X {method}".format(method=method)

    def get_item_content(self, clear_password=False):
        get_cmd = " ".join(
            [
                self.get_prefix("GET"),
                self.endpoint + "/" + self.test_item_name,
                "-d output_mode=json",
            ]
        )
        if clear_password:
            get_cmd = " ".join([get_cmd, "-d --cred--=1"])
        response = subprocess.check_output(get_cmd.split(" "))
        try:
            item = json.loads(response)
        except ValueError:
            self.fail("Could not get the item %s" % self.test_item_name)

        # get the item content
        try:
            content = item["entry"][0]["content"]
        except KeyError:
            self.fail("Could not get the item content")
        return content

    def get_all_item(self, clear_password=False):
        get_all_cmd = " ".join(
            [self.get_prefix("GET"), self.endpoint, "-d output_mode=json"]
        )
        if clear_password:
            get_all_cmd = " ".join([get_all_cmd, "-d --cred--=1"])
        response = subprocess.check_output(get_all_cmd.split(" "))
        try:
            items = json.loads(response)
        except ValueError:
            self.fail("Could not get the item %s" % self.test_item_name)
        return items

    def get_all_passwords(self):
        get_password_cmd = " ".join(
            [self.get_prefix("GET"), self.password_endpoint, "-d output_mode=json"]
        )
        output = subprocess.check_output(get_password_cmd.split(" "))
        try:
            passwords = json.loads(output)
        except ValueError:
            self.fail("Could not get the passwords")
        return passwords

    def testCreateWithExisting(self):
        """
        Test: Create an item which already exists
        """
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )

        # create an item with the same name
        create_cmd = " ".join(
            [
                self.get_prefix("POST"),
                self.endpoint,
                "-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                    name=self.test_item_name
                ),
            ]
        )
        output = subprocess.check_output(create_cmd.split(" "))
        # test the error message
        self.assertTrue(
            'REST Error [409]: Conflict -- Name "{name}" is already in use'.format(
                name=self.test_item_name
            )
            in output
        )

        self.delete_item()

    def testCreateWithEmptyOptionalField(self):
        """
        Test: Create an item with empty optional field which has defaultValue defined in schema

        """
        # create an item with the empty field value
        # defaultValue is defined with '123456789' in schema
        self.create_item(
            data="-d name={name} -d api_uuid= -d endpoint=1 -d api_key=1".format(
                name=self.test_item_name
            )
        )

        content = self.get_item_content()
        # test the empty field 'api_uuid' does not exist in content
        self.assertFalse("api_uuid" in content)
        self.delete_item()

    def testCreateWithMultipleEmptyOptionalField(self):
        """
        Test: Create an item with empty optional field which has defaultValue defined in schema

        """
        # create an item with multiple empty field value
        # defaultValue is defined with '123456789' in schema for both 'api_uuid' and 'api_uuid2'
        self.create_item(
            data="-d name={name} -d api_uuid= -d api_uuid2= -d endpoint=1 -d api_key=1".format(
                name=self.test_item_name
            )
        )

        content = self.get_item_content()
        # test the empty field 'api_uuid' and 'api_uuid2' does not exist in content
        self.assertFalse("api_uuid" in content)
        self.assertFalse("api_uuid2" in content)
        self.delete_item()

    def testCreateWithoutOptionalField(self):
        """
        Test: Create an item without optional field which has defaultValue defined in schema

        """
        # create an item without optional field 'api_uuid'
        # defaultValue is defined with '123456789' in schema
        self.create_item(
            data="-d name={name} -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )

        content = self.get_item_content()
        # test the empty field 'api_uuid' does not exist in content
        self.assertFalse("api_uuid" in content)
        self.delete_item()

    def testCreateWithMagicPassword(self):
        """
        Test: Create an item with '******' as password value

        """
        # create item with '*****' as value
        self.create_item(
            data="-d name={name} -d endpoint=1 -d api_key=****** -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )

        content = self.get_item_content(clear_password=True)

        # check if the password equals the magic password
        self.assertEqual(content["api_key"], self.PASSWORD)
        self.delete_item()

    def testCreateWithMultipleMagicPassword(self):
        """
        Test: Create an item with '******' as password value

        """
        # create item with '*******' as value
        self.create_item(
            data="-d name={name} -d endpoint=1 -d api_key=****** -d test_optional_password=******".format(
                name=self.test_item_name
            )
        )

        content = self.get_item_content(clear_password=True)

        # check if the password equals the magic password
        self.assertEqual(content["api_key"], self.PASSWORD)
        self.assertEqual(content["test_optional_password"], self.PASSWORD)
        self.delete_item()

    def testCreateWithEmptyRequiredField(self):
        """
        Test: Create an account with empty required field 'api_key'
        """
        create_cmd = " ".join(
            [
                self.get_prefix("POST"),
                self.endpoint,
                "-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=".format(
                    name=self.test_item_name
                ),
            ]
        )
        output = subprocess.check_output(create_cmd.split(" "))
        self.assertTrue("Required field is missing: api_key" in output)

    def testCreateWithNoneEmptyRequiredField(self):
        # create an account with:
        # 1.none empty required field 'api_key'
        # 2.no field 'test_optional_password'
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=".format(
                name=self.test_item_name
            )
        )

        item_content = self.get_item_content()
        # check the get request response contains only 'api_key=******'
        self.assertEqual(item_content["api_key"], self.PASSWORD)
        # check the get request response does not contain test_optional_password field
        self.assertFalse("test_optional_password" in item_content)

        clear_item_content = self.get_item_content(clear_password=True)
        # check the clear password is stored in passwords.conf
        self.assertEqual(clear_item_content["api_key"], self.test_item_name)
        # check the get request response does not contain test_optional_password field
        self.assertFalse("test_optional_password" in clear_item_content)

        # delete created item
        self.delete_item()

    def testUpdateNoneExistedItem(self):
        """
        Test: Update an item which does exist
        """
        data = "-d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1"
        update_cmd = " ".join(
            [self.get_prefix("POST"), self.endpoint + "/" + self.test_item_name, data]
        )
        output = subprocess.check_output(update_cmd.split(" "))
        self.assertTrue(
            'REST Error [404]: Not Found -- "{name}" does not exist'.format(
                name=self.test_item_name
            )
            in output
        )

    def testUpdateWithEmptyValueForRequiredField(self):
        """
        Test: Update an item with empty value for required field
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )
        update_cmd = " ".join(
            [
                self.get_prefix("POST"),
                self.endpoint + "/" + self.test_item_name,
                "-d api_uuid=1 -d endpoint=1 -d api_key=",
            ]
        )
        output = subprocess.check_output(update_cmd.split(" "))
        self.assertTrue("Required field is missing: api_key" in output)
        self.delete_item()

    def testUpdateWithMagicPassword(self):
        """
        Test: Update existing password with '******', the password should not change after update
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )
        # update with password = '******'
        self.update_item(data="-d api_key=******")

        clear_item_content = self.get_item_content(clear_password=True)
        # check the clear password is unchanged
        self.assertEqual(clear_item_content["api_key"], "1")

        # delete the created item
        self.delete_item()

    def testUpdateWithOtherPassword(self):
        """
        Test: Update existing password field value '1' with value 'other'.
        The password field should change to 'other'
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )

        # update the test item
        self.update_item(data="-d api_key=other")

        clear_item_content = self.get_item_content(clear_password=True)
        # check the clear password is unchanged
        self.assertEqual(clear_item_content["api_key"], "other")

        # delete the created item
        self.delete_item()

    def testUpdateWithMultipleEmptyFieldValue(self):
        """
        Test: Create item with f1=1 and f2=2, update item with f1='' and f2=''
        The update should empty field in conf as:
            f1 =
            f2 =

        case 5 and case 6
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d f1=1 -d f2=2".format(
                name=self.test_item_name
            )
        )
        # update with f1='' and f2=''
        self.update_item(data="-d endpoint=1 -d api_key=1 -d f1= -d f2=")
        content = self.get_item_content()
        self.assertEqual(content["f1"], "")
        self.assertEqual(content["f2"], "")

        # update with f1='1' and f2='2'
        self.update_item(data="-d endpoint=1 -d api_key=1 -d f1=1 -d f2=2")
        content = self.get_item_content(clear_password=True)
        self.assertEqual(content["f1"], "1")
        self.assertEqual(content["f2"], "2")

        self.delete_item()

    def testUpdateWithAdditionalFields(self):
        """
        Test: Create item with 'f1=1', update it with 'f1=2, f2=2, f3=3'
        New field values should be added to the item

        case 7: update + add
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d f1=1".format(
                name=self.test_item_name
            )
        )
        self.update_item(data="-d endpoint=1 -d api_key=1 -d f1=1 -d f2=2 -d f3=3")

        content = self.get_item_content(clear_password=True)
        self.assertEqual(content["f1"], "1")
        self.assertEqual(content["f2"], "2")
        self.assertEqual(content["f3"], "3")

        self.delete_item()

    def testUpdateWithOneEmptyFieldValue(self):
        """
        Test: Create an item with 'f1=1, f2=2, f3=3', update it with 'f1=f1, f2=f2, f3='
        field f3 should be empty after update
        case 8: update + remove
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d f1=1 -d f2=2 -d f3=3".format(
                name=self.test_item_name
            )
        )
        self.update_item(data="-d endpoint=1 -d api_key=1 -d f1=f1 -d f2=f2 -d f3=")

        content = self.get_item_content(clear_password=True)
        self.assertEqual(content["f1"], "f1")
        self.assertEqual(content["f2"], "f2")
        self.assertEqual(content["f3"], "")

        self.delete_item()

    def testUpdateWithPartialFields(self):
        """
        Test: Create an item with 'f1=1, f2=2, f3=3', update it with 'f1=f1, f2='
        The update result should be 'f1=f1, f2=, f3=3'
        case 9: partial update
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d f1=1 -d f2=2 -d f3=3".format(
                name=self.test_item_name
            )
        )
        self.update_item(data="-d endpoint=1 -d api_key=1 -d f1=f1 -d f2=")

        content = self.get_item_content(clear_password=True)
        self.assertEqual(content["f1"], "f1")
        self.assertEqual(content["f2"], "")
        # keep original password if it's not passed
        self.assertEqual(content["f3"], "3")

        self.delete_item()

    def testUpdateWithMixedUpdate(self):
        """
        Test: Create an item with 'f1=1, f2=2, f3=3, f4=4'
        Update it with 'f1=f1, f2=, f3=f3, f5=f5'
        The update result should be 'f1=f1, f2=, f3=f3, f4=4 f5=f5'
        case 10: mixed update
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d f1=1 -d f2=2 -d f3=3 -d f4=4".format(
                name=self.test_item_name
            )
        )
        self.update_item(
            data="-d endpoint=1 -d api_key=1 -d f1=f1 -d f2= -d f3=f3 -d f5=f5"
        )

        content = self.get_item_content(clear_password=True)
        # changed
        self.assertEqual(content["f1"], "f1")
        # reset
        self.assertEqual(content["f2"], "")
        # changed
        self.assertEqual(content["f3"], "f3")
        # unchanged
        self.assertEqual(content["f4"], "4")
        # added
        self.assertEqual(content["f5"], "f5")

        self.delete_item()

    def testUpdateWithMixedRemove(self):
        """
        Test: Create an item with 'f1=1, f2=2, f3=3, f4=4'
        Update it with 'f1=, f2=, f3=, f5='
        The update result should be 'f1=, f2=, f3=f3, f4=4, f5='
        case 11: mixed remove
        """
        # create a test item
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d f1=1 -d f2=2 -d f3=3 -d f4=4".format(
                name=self.test_item_name
            )
        )
        self.update_item(data="-d endpoint=1 -d api_key=1 -d f1= -d f2= -d f3= -d f5=")

        content = self.get_item_content(clear_password=True)
        self.assertEqual(content["f1"], "")
        self.assertEqual(content["f2"], "")
        self.assertEqual(content["f3"], "")
        self.assertEqual(content["f4"], "4")
        self.assertFalse("f5" in content)

        self.delete_item()

    def testGetEncrypt(self):
        """
        Test: Create a stanza via conf, the password fields are encrypted when calling get REST call
        """
        # create account via conf
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            ),
            via_conf=True,
        )

        content = self.get_item_content()
        self.assertEqual(content["api_key"], self.PASSWORD)
        self.assertEqual(content["test_optional_password"], self.PASSWORD)

        clear_content = self.get_item_content(clear_password=True)
        self.assertEqual(clear_content["api_key"], "1")
        self.assertEqual(clear_content["test_optional_password"], "1")

        self.delete_item()

    def testGetWithOnePasswordChanged(self):
        """
        Test: Change one password field via conf and get via REST call, the change takes effect

        """
        # create account with password encrypted
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )
        # update password with other value via conf
        self.update_item(
            data="-d api_key=other -d test_optional_password=other", via_conf=True
        )

        # get account
        content = self.get_item_content()
        self.assertEqual(content["api_key"], self.PASSWORD)
        self.assertEqual(content["test_optional_password"], self.PASSWORD)

        # get account with clear password option
        clear_content = self.get_item_content(clear_password=True)
        self.assertEqual(clear_content["api_key"], "other")
        self.assertEqual(clear_content["test_optional_password"], "other")

        self.delete_item()

    def testGetWithEmptyOptionalPassword(self):
        """
        Test: Change optional password field to empty in conf
              Get the stanza via REST will get empty string in response
        """
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )

        # update password with empty password value via conf
        self.update_item(data="-d api_key=1 -d test_optional_password=", via_conf=True)

        # get account
        content = self.get_item_content()
        self.assertEqual(content["api_key"], self.PASSWORD)
        self.assertEqual(content["test_optional_password"], "")

        # get account with clear password option
        clear_content = self.get_item_content(clear_password=True)
        self.assertEqual(clear_content["api_key"], "1")
        self.assertEqual(clear_content["test_optional_password"], "")

        self.delete_item()

    def testAllWithClearPassword(self):
        """
        Test: Create two stanzas via conf with clear passwords api_key=1 and test_optional_password=1
        Get all stanzas via REST handler. The clear passwords should be encrypted in passwords.conf
        The password fields should be '*******' in response

        """
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name="stanza1"
            ),
            via_conf=True,
            name="stanza1",
        )
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=2 -d test_optional_password=2".format(
                name="stanza2"
            ),
            via_conf=True,
            name="stanza2",
        )
        items = self.get_all_item()
        for item in items["entry"]:
            self.assertEqual(item["content"]["api_key"], self.PASSWORD)
            self.assertEqual(item["content"]["test_optional_password"], self.PASSWORD)

        items_with_clear_password = self.get_all_item(clear_password=True)
        for item in items_with_clear_password["entry"]:
            if item["name"] == "stanza1":
                self.assertEqual(item["content"]["api_key"], "1")
                self.assertEqual(item["content"]["test_optional_password"], "1")
            elif item["name"] == "stanza2":
                self.assertEqual(item["content"]["api_key"], "2")
                self.assertEqual(item["content"]["test_optional_password"], "2")

        self.delete_item(name="stanza1")
        self.delete_item(name="stanza2")

    def testAllWithEmptyField(self):
        """
        Test:
        1.Create two stanzas via conf with clear passwords:
            stanza1: api_key=1
            stanza2: api_key=2 and test_optional_password=2
        2.Update stanza2 with: test_optional_password=
        3.Get all stanzas via REST handler. The clear passwords should be encrypted in passwords.conf
            For stanza1, test_optional_password should not be in response.
            For stanza2, test_optional_password='' in response
        """
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1".format(
                name="stanza1"
            ),
            via_conf=True,
            name="stanza1",
        )
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=2 -d test_optional_password=2".format(
                name="stanza2"
            ),
            via_conf=True,
            name="stanza2",
        )

        self.update_item(
            data="-d api_key=2 -d endpoint=1 -d test_optional_password=",
            via_conf=True,
            name="stanza2",
        )
        items = self.get_all_item()
        for item in items["entry"]:
            if item["name"] == "stanza1":
                self.assertEqual(item["content"]["api_key"], self.PASSWORD)
                self.assertFalse("test_optional_password" in item["content"])
            if item["name"] == "stanza2":
                self.assertEqual(item["content"]["api_key"], self.PASSWORD)
                self.assertEqual(item["content"]["test_optional_password"], "")

        items_with_clear_password = self.get_all_item(clear_password=True)
        for item in items_with_clear_password["entry"]:
            if item["name"] == "stanza1":
                self.assertEqual(item["content"]["api_key"], "1")
                self.assertFalse("test_optional_password" in item["content"])
            elif item["name"] == "stanza2":
                self.assertEqual(item["content"]["api_key"], "2")
                self.assertEqual(item["content"]["test_optional_password"], "")

        self.delete_item(name="stanza1")
        self.delete_item(name="stanza2")

    def testDeleteNoneExistentItem(self):
        """
        Test: Delete item that does not exist
              Current behaviour is getting error response
        """
        delete_cmd = " ".join(
            [
                self.get_prefix("DELETE"),
                self.endpoint + "/" + self.test_item_name,
                "-d output_mode=json",
            ]
        )
        output = subprocess.check_output(delete_cmd.split(" "))
        self.assertTrue(
            "Could not find object id={name}".format(name=self.test_item_name) in output
        )

    def testDeleteItem(self):
        """
        Test: Delete an item with password. The encrypted fields in passwords.conf should be deleted too.
        """
        realm = "#Splunk_TA_UCCExample#configs/conf-splunk_ta_uccexample_account"
        passwords = self.get_all_passwords()

        for password in passwords["entry"]:
            if realm in password["name"]:
                self.fail("Password should not exist in passwords.conf")
        self.create_item(
            data="-d name={name} -d api_uuid=1 -d endpoint=1 -d api_key=1 -d test_optional_password=1".format(
                name=self.test_item_name
            )
        )
        # get all passwords again
        passwords = self.get_all_passwords()

        for password in passwords["entry"]:
            if realm in password["name"]:
                break
        else:
            self.fail("Password should exist in passwords.conf")

        self.delete_item()

        # get passwords after deleting the item
        passwords = self.get_all_passwords()

        for password in passwords["entry"]:
            if realm in password["name"]:
                self.fail("Password should not exist in passwords.conf")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['splunktalib',
 'splunktalib.common',
 'splunktalib.concurrent',
 'splunktalib.conf_manager',
 'splunktalib.schedule',
 'splunktalib.tests',
 'splunktalib.tests.splunktaucclib.functional']

package_data = \
{'': ['*'],
 'splunktalib.tests': ['data/*', 'data/package/*', 'data/package/default/*']}

install_requires = \
['httplib2>=0.18,<0.19', 'sortedcontainers>=2.2,<3.0']

setup_kwargs = {
    'name': 'splunktalib',
    'version': '1.1.5a1',
    'description': 'Supporting library for Splunk Add-ons',
    'long_description': None,
    'author': 'rfaircloth-splunk',
    'author_email': 'rfaircloth@splunk.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/splunk/addonfactory-ta-library-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*',
}


setup(**setup_kwargs)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

import re
from xml.etree import cElementTree as et


def parse_conf_xml_dom(xml_content):
    """
    @xml_content: XML DOM from splunkd
    """
    xml_content = xml_content.decode("utf-8")
    m = re.search(r'xmlns="([^"]+)"', xml_content)
    ns = m.group(1)
    m = re.search(r'xmlns:s="([^"]+)"', xml_content)
    sub_ns = m.group(1)
    entry_path = "./{%s}entry" % ns
    stanza_path = "./{%s}title" % ns
    key_path = "./{%s}content/{%s}dict/{%s}key" % (ns, sub_ns, sub_ns)
    meta_path = "./{%s}dict/{%s}key" % (sub_ns, sub_ns)
    list_path = "./{%s}list/{%s}item" % (sub_ns, sub_ns)

    xml_conf = et.fromstring(xml_content)
    stanza_objs = []
    for entry in xml_conf.iterfind(entry_path):
        for stanza in entry.iterfind(stanza_path):
            stanza_obj = {"name": stanza.text, "stanza": stanza.text}
            break
        else:
            continue

        for key in entry.iterfind(key_path):
            if key.get("name") == "eai:acl":
                meta = {}
                for k in key.iterfind(meta_path):
                    meta[k.get("name")] = k.text
                stanza_obj[key.get("name")] = meta
            elif key.get("name") != "eai:attributes":
                name = key.get("name")
                if name.startswith("eai:"):
                    name = name[4:]
                list_vals = [k.text for k in key.iterfind(list_path)]
                if list_vals:
                    stanza_obj[name] = list_vals
                else:
                    stanza_obj[name] = key.text
                    if key.text == "None":
                        stanza_obj[name] = None
        stanza_objs.append(stanza_obj)
    return stanza_objs



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
A simple thread pool implementation
"""
from __future__ import division

from future import standard_library

standard_library.install_aliases()
from builtins import range
from builtins import object
import threading
import queue
import multiprocessing
import traceback
from time import time
from splunktalib.common import log


class ThreadPool(object):
    """
    A simple thread pool implementation
    """

    _high_watermark = 0.2
    _resize_window = 10

    def __init__(self, min_size=1, max_size=128, task_queue_size=1024, daemon=True):
        assert task_queue_size

        if not min_size or min_size <= 0:
            min_size = multiprocessing.cpu_count()

        if not max_size or max_size <= 0:
            max_size = multiprocessing.cpu_count() * 8

        self._min_size = min_size
        self._max_size = max_size
        self._daemon = daemon

        self._work_queue = queue.Queue(task_queue_size)
        self._thrs = []
        for _ in range(min_size):
            thr = threading.Thread(target=self._run)
            self._thrs.append(thr)
        self._admin_queue = queue.Queue()
        self._admin_thr = threading.Thread(target=self._do_admin)
        self._last_resize_time = time()
        self._last_size = min_size
        self._lock = threading.Lock()
        self._occupied_threads = 0
        self._count_lock = threading.Lock()
        self._started = False

    def start(self):
        """
        Start threads in the pool
        """

        with self._lock:
            if self._started:
                return
            self._started = True

            for thr in self._thrs:
                thr.daemon = self._daemon
                thr.start()

            self._admin_thr.start()
        log.logger.info("ThreadPool started.")

    def tear_down(self):
        """
        Tear down thread pool
        """

        with self._lock:
            if not self._started:
                return
            self._started = False

            for thr in self._thrs:
                self._work_queue.put(None, block=True)

            self._admin_queue.put(None)

            if not self._daemon:
                log.logger.info("Wait for threads to stop.")
                for thr in self._thrs:
                    thr.join()
            self._admin_thr.join()

        log.logger.info("ThreadPool stopped.")

    def enqueue_funcs(self, funcs, block=True):
        """
        run jobs in a fire and forget way, no result will be handled
        over to clients
        :param funcs: tuple/list-like or generator like object, func shall be
                      callable
        """

        if not self._started:
            log.logger.info("ThreadPool has already stopped.")
            return

        for func in funcs:
            self._work_queue.put(func, block)

    def apply_async(self, func, args=(), kwargs=None, callback=None):
        """
        :param func: callable
        :param args: free params
        :param kwargs: named params
        :callback: when func is done and without exception, call the callback
        :return AsyncResult, clients can poll or wait the result through it
        """

        if not self._started:
            log.logger.info("ThreadPool has already stopped.")
            return None

        res = AsyncResult(func, args, kwargs, callback)
        self._work_queue.put(res)
        return res

    def apply(self, func, args=(), kwargs=None):
        """
        :param func: callable
        :param args: free params
        :param kwargs: named params
        :return whatever the func returns
        """

        if not self._started:
            log.logger.info("ThreadPool has already stopped.")
            return None

        res = self.apply_async(func, args, kwargs)
        return res.get()

    def size(self):
        return self._last_size

    def resize(self, new_size):
        """
        Resize the pool size, spawn or destroy threads if necessary
        """

        if new_size <= 0:
            return

        if self._lock.locked() or not self._started:
            log.logger.info(
                "Try to resize thread pool during the tear " "down process, do nothing"
            )
            return

        with self._lock:
            self._remove_exited_threads_with_lock()
            size = self._last_size
            self._last_size = new_size
            if new_size > size:
                for _ in range(new_size - size):
                    thr = threading.Thread(target=self._run)
                    thr.daemon = self._daemon
                    thr.start()
                    self._thrs.append(thr)
            elif new_size < size:
                for _ in range(size - new_size):
                    self._work_queue.put(None)
        log.logger.info("Finished ThreadPool resizing. New size=%d", new_size)

    def _remove_exited_threads_with_lock(self):
        """
        Join the exited threads last time when resize was called
        """

        joined_thrs = set()
        for thr in self._thrs:
            if not thr.is_alive():
                try:
                    if not thr.daemon:
                        thr.join(timeout=0.5)
                    joined_thrs.add(thr.ident)
                except RuntimeError:
                    pass

        if joined_thrs:
            live_thrs = []
            for thr in self._thrs:
                if thr.ident not in joined_thrs:
                    live_thrs.append(thr)
            self._thrs = live_thrs

    def _do_resize_according_to_loads(self):
        if (
            self._last_resize_time
            and time() - self._last_resize_time < self._resize_window
        ):
            return

        thr_size = self._last_size
        free_thrs = thr_size - self._occupied_threads
        work_size = self._work_queue.qsize()

        log.logger.debug(
            "current_thr_size=%s, free_thrs=%s, work_size=%s",
            thr_size,
            free_thrs,
            work_size,
        )
        if work_size and work_size > free_thrs:
            if thr_size < self._max_size:
                thr_size = min(thr_size * 2, self._max_size)
                self.resize(thr_size)
        elif free_thrs > 0:
            free = free_thrs * 1.0
            if free / thr_size >= self._high_watermark and free_thrs >= 2:
                # 20 % thrs are idle, tear down half of the idle ones
                thr_size = thr_size - int(free_thrs // 2)
                if thr_size > self._min_size:
                    self.resize(thr_size)
        self._last_resize_time = time()

    def _do_admin(self):
        admin_q = self._admin_queue
        resize_win = self._resize_window
        while 1:
            try:
                wakup = admin_q.get(timeout=resize_win + 1)
            except queue.Empty:
                self._do_resize_according_to_loads()
                continue

            if wakup is None:
                break
            else:
                self._do_resize_according_to_loads()
        log.logger.info(
            "ThreadPool admin thread=%s stopped.", threading.current_thread().getName()
        )

    def _run(self):
        """
        Threads callback func, run forever to handle jobs from the job queue
        """

        work_queue = self._work_queue
        count_lock = self._count_lock
        while 1:
            log.logger.debug("Going to get job")
            func = work_queue.get()
            if func is None:
                break

            if not self._started:
                break

            log.logger.debug("Going to exec job")
            with count_lock:
                self._occupied_threads += 1

            try:
                func()
            except Exception:
                log.logger.error(traceback.format_exc())

            with count_lock:
                self._occupied_threads -= 1

            log.logger.debug("Done with exec job")
            log.logger.info("Thread work_queue_size=%d", work_queue.qsize())

        log.logger.debug(
            "Worker thread %s stopped.", threading.current_thread().getName()
        )


class AsyncResult(object):
    def __init__(self, func, args, kwargs, callback):
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self._callback = callback
        self._q = queue.Queue()

    def __call__(self):
        try:
            if self._args and self._kwargs:
                res = self._func(*self._args, **self._kwargs)
            elif self._args:
                res = self._func(*self._args)
            elif self._kwargs:
                res = self._func(**self._kwargs)
            else:
                res = self._func()
        except Exception as e:
            self._q.put(e)
            return
        else:
            self._q.put(res)

        if self._callback is not None:
            self._callback()

    def get(self, timeout=None):
        """
        Return the result when it arrives. If timeout is not None and the
        result does not arrive within timeout seconds then
        multiprocessing.TimeoutError is raised. If the remote call raised an
        exception then that exception will be reraised by get().
        """

        try:
            res = self._q.get(timeout=timeout)
        except queue.Empty:
            raise multiprocessing.TimeoutError("Timed out")

        if isinstance(res, Exception):
            raise res
        return res

    def wait(self, timeout=None):
        """
        Wait until the result is available or until timeout seconds pass.
        """

        try:
            res = self._q.get(timeout=timeout)
        except queue.Empty:
            pass
        else:
            self._q.put(res)

    def ready(self):
        """
        Return whether the call has completed.
        """

        return len(self._q)

    def successful(self):
        """
        Return whether the call completed without raising an exception.
        Will raise AssertionError if the result is not ready.
        """

        if not self.ready():
            raise AssertionError("Function is not ready")
        res = self._q.get()
        self._q.put(res)

        if isinstance(res, Exception):
            return False
        return True



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

import sys
import subprocess
import traceback

import splunktalib.splunk_platform as sp
from splunktalib.common import log


def _parse_modinput_configs(root, outer_block, inner_block):
    """
    When user splunkd spawns modinput script to do config check or run

    <?xml version="1.0" encoding="UTF-8"?>
    <input>
      <server_host>localhost.localdomain</server_host>
      <server_uri>https://127.0.0.1:8089</server_uri>
      <session_key>xxxyyyzzz</session_key>
      <checkpoint_dir>ckpt_dir</checkpoint_dir>
      <configuration>
        <stanza name="snow://alm_asset">
          <param name="duration">60</param>
            <param name="host">localhost.localdomain</param>
            <param name="index">snow</param>
            <param name="priority">10</param>
        </stanza>
        ...
      </configuration>
    </input>

    When user create an stanza through data input on WebUI

    <?xml version="1.0" encoding="UTF-8"?>
    <items>
      <server_host>localhost.localdomain</server_host>
      <server_uri>https://127.0.0.1:8089</server_uri>
      <session_key>xxxyyyzzz</session_key>
      <checkpoint_dir>ckpt_dir</checkpoint_dir>
      <item name="abc">
        <param name="duration">60</param>
        <param name="exclude"></param>
        <param name="host">localhost.localdomain</param>
        <param name="index">snow</param>
        <param name="priority">10</param>
      </item>
    </items>
    """

    confs = root.getElementsByTagName(outer_block)
    if not confs:
        log.logger.error("Invalid config, missing %s section", outer_block)
        raise Exception("Invalid config, missing %s section".format(outer_block))

    configs = []
    stanzas = confs[0].getElementsByTagName(inner_block)
    for stanza in stanzas:
        config = {}
        stanza_name = stanza.getAttribute("name")
        if not stanza_name:
            log.logger.error("Invalid config, missing name")
            raise Exception("Invalid config, missing name")

        config["name"] = stanza_name
        params = stanza.getElementsByTagName("param")
        for param in params:
            name = param.getAttribute("name")
            if (
                name
                and param.firstChild
                and param.firstChild.nodeType == param.firstChild.TEXT_NODE
            ):
                config[name] = param.firstChild.data
        configs.append(config)
    return configs


def parse_modinput_configs(config_str):
    """
    @config_str: modinput XML configuration feed by splunkd
    @return: meta_config and stanza_config
    """

    import xml.dom.minidom as xdm

    meta_configs = {
        "server_host": None,
        "server_uri": None,
        "session_key": None,
        "checkpoint_dir": None,
    }
    root = xdm.parseString(config_str)
    doc = root.documentElement
    for tag in meta_configs.keys():
        nodes = doc.getElementsByTagName(tag)
        if not nodes:
            log.logger.error("Invalid config, missing %s section", tag)
            raise Exception("Invalid config, missing %s section", tag)

        if nodes[0].firstChild and nodes[0].firstChild.nodeType == nodes[0].TEXT_NODE:
            meta_configs[tag] = nodes[0].firstChild.data
        else:
            log.logger.error("Invalid config, expect text ndoe")
            raise Exception("Invalid config, expect text ndoe")

    if doc.nodeName == "input":
        configs = _parse_modinput_configs(doc, "configuration", "stanza")
    else:
        configs = _parse_modinput_configs(root, "items", "item")
    return meta_configs, configs


def get_modinput_configs_from_cli(modinput, modinput_stanza=None):
    """
    @modinput: modinput name
    @modinput_stanza: modinput stanza name, for multiple instance only
    """

    assert modinput

    splunkbin = sp.get_splunk_bin()
    cli = [splunkbin, "cmd", "splunkd", "print-modinput-config", modinput]
    if modinput_stanza:
        cli.append(modinput_stanza)

    out, err = subprocess.Popen(
        cli, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ).communicate()
    if err:
        log.logger.error("Failed to get modinput configs with error: %s", err)
        return None, None
    else:
        return parse_modinput_configs(out)


def get_modinput_config_str_from_stdin():
    """
    Get modinput from stdin which is feed by splunkd
    """

    try:
        return sys.stdin.read(5000)
    except Exception:
        log.logger.error(traceback.format_exc())
        raise


def get_modinput_configs_from_stdin():
    config_str = get_modinput_config_str_from_stdin()
    return parse_modinput_configs(config_str)



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

from builtins import range
from builtins import object
import re
import json
from xml.etree import cElementTree as et

import splunktalib.rest as rest


class KVException(Exception):
    pass


class KVAlreadyExists(KVException):
    pass


class KVNotExists(KVException):
    pass


class KVClient(object):
    def __init__(self, splunkd_host, session_key):
        self._splunkd_host = splunkd_host
        self._session_key = session_key

    def create_collection(self, collection, app, owner="nobody"):
        """
        :collection: collection name
        :return: None if successful otherwise KV exception thrown
        """

        assert collection
        assert app

        uri = self._get_config_endpoint(app, owner)
        data = {"name": collection}
        self._do_request(uri, "POST", data)

    def list_collection(self, collection=None, app=None, owner="nobody"):
        """
        :collection: collection name. When euqals "None", return all
        collections in the system.
        :return: a list containing the connection names if successful, throws
        KVNotExists if no such colection or other exception if other error
        happened
        """

        uri = self._get_config_endpoint(app, owner, collection)

        content = self._do_request(uri, method="GET")
        m = re.search(r'xmlns="([^"]+)"', content)
        path = "./entry/title"
        if m:
            ns = m.group(1)
            path = "./{%s}entry/{%s}title" % (ns, ns)

        collections = et.fromstring(content)
        return [node.text for node in collections.iterfind(path)]

    def delete_collection(self, collection, app, owner="nobody"):
        """
        :collection: collection name to be deleted
        :return: None if successful otherwise throw KVNotExists exception if
        the collection doesn't exist in the system or other exception if other
        error happened
        """

        assert collection

        uri = self._get_config_endpoint(app, owner, collection)
        self._do_request(uri, method="DELETE")

    def insert_collection_data(self, collection, data, app, owner="nobody"):
        """
        :collection: collection name
        :data: dict like key values to be inserted and attached to
        this collection
        :return: {"_key": "key_id"} when successful, clients can use this
        key to do query/delete/update, throws KV exceptions when failed
        """

        assert collection
        assert data is not None
        assert app

        uri = self._get_data_endpoint(app, owner, collection)
        key = self._do_request(uri, "POST", data, content_type="application/json")
        return json.loads(key)

    def delete_collection_data(self, collection, key_id, app, owner="nobody"):
        """
        :collection: collection name
        :key_id: key id returned when creation. If None, delete all data
        associated with this collection
        :return: None if successful otherwise throws KV exception
        """

        assert collection

        uri = self._get_data_endpoint(app, owner, collection, key_id)
        self._do_request(uri, "DELETE", content_type="application/json")

    def update_collection_data(self, collection, key_id, data, app, owner="nobody"):
        """
        :collection: collection name
        :key_id: key id returned when creation
        :return: key id if successful otherwise throws KV exception
        """

        assert collection
        assert key_id
        assert app

        uri = self._get_data_endpoint(app, owner, collection, key_id)
        k = self._do_request(uri, "POST", data, content_type="application/json")
        return json.loads(k)

    def get_collection_data(self, collection, key_id, app, owner="nobody"):
        """
        :collection: collection name
        :key_id: key id returned when creation. If None, get all data
        associated with this collection
        :return: when key_id is not None, return key values if
        successful. when key_id is None, return a list of key values if
        sucessful. Throws KV exception if failure
        """

        assert collection

        uri = self._get_data_endpoint(app, owner, collection, key_id)
        k = self._do_request(uri, "GET")
        return json.loads(k)

    def _do_request(
        self, uri, method, data=None, content_type="application/x-www-form-urlencoded"
    ):
        headers = {"Content-Type": content_type}

        resp, content = rest.splunkd_request(
            uri, self._session_key, method, headers, data
        )
        if resp is None and content is None:
            raise KVException("Failed uri={0}, data={1}".format(uri, data))

        if resp.status in (200, 201):
            return content
        elif resp.status == 409:
            raise KVAlreadyExists("{0}-{1} already exists".format(uri, data))
        elif resp.status == 404:
            raise KVNotExists("{0}-{1} not exists".format(uri, data))
        else:
            raise KVException(
                "Failed to {0} {1}, reason={2}".format(method, uri, resp.reason)
            )

    def _get_config_endpoint(self, app, owner, collection=None):
        uri = "{0}/servicesNS/{1}/{2}/storage/collections/config"
        return self._do_get_endpoint(app, owner, collection, None, uri)

    def _get_data_endpoint(self, app, owner, collection, key_id=None):
        uri = "{0}/servicesNS/{1}/{2}/storage/collections/data"
        return self._do_get_endpoint(app, owner, collection, key_id, uri)

    def _do_get_endpoint(self, app, owner, collection, key_id, uri_template):
        if not app:
            app = "-"

        if not owner:
            owner = "-"

        uri = uri_template.format(self._splunkd_host, owner, app)

        if collection is not None:
            uri += "/{0}".format(collection)
            if key_id is not None:
                uri += "/{0}".format(key_id)
        return uri


def create_collection(kv_client, collection, appname):
    not_exists = False
    try:
        res = kv_client.list_collection(collection, appname)
    except KVNotExists:
        not_exists = True
    except Exception:
        not_exists = True

    if not_exists or not res:
        for i in range(3):
            try:
                kv_client.create_collection(collection, appname)
            except KVAlreadyExists:
                return
            except Exception as e:
                ex = e
            else:
                return
        else:
            raise ex



# SPDX-FileCopyrightText: 2020 Splunk Inc.
#
# SPDX-License-Identifier: Apache-2.0

"""
Handles credentials related stuff
"""

from builtins import str
from builtins import range
from builtins import object
import re
import xml.dom.minidom as xdm

import splunktalib.common.util as util
import splunktalib.common.xml_dom_parser as xdp
import splunktalib.rest as rest

# Splunk can only encrypt string when length <=255
SPLUNK_CRED_LEN_LIMIT = 255


class CredException(Exception):
    pass


class CredNotFound(CredException):
    """
    Credential information not exists
    """

    pass


def create_credential_manager(username, password, splunkd_uri, app, owner, realm):
    session_key = CredentialManager.get_session_key(username, password, splunkd_uri)
    return CredentialManager(splunkd_uri, session_key, app, owner, realm)


class CredentialManager(object):
    """
    Credential related interfaces
    """

    def __init__(self, splunkd_uri, session_key, app="-", owner="nobody", realm=None):
        """
        :app: when creating/upating/deleting app is required
        """

        self._app = app
        self._splunkd_uri = splunkd_uri
        self._owner = owner
        self._sep = "``splunk_cred_sep``"

        if realm:
            self._realm = realm
        else:
            self._realm = app

        self._session_key = session_key

    def set_appname(self, app):
        """
        This are cases we need edit/remove/create confs in different app
        context. call this interface to switch app context before manipulate
        the confs in different app context
        """

        self._app = app

    @staticmethod
    def get_session_key(username, password, splunkd_uri="https://localhost:8089"):
        """
        Get session key by using login username and passwrod
        :return: session_key if successful, None if failed
        """

        eid = "".join((splunkd_uri, "/services/auth/login"))
        postargs = {
            "username": username,
            "password": password,
        }

        response, content = rest.splunkd_request(
            eid, None, method="POST", data=postargs
        )

        if response is None and content is None:
            raise CredException("Get session key failed.")

        xml_obj = xdm.parseString(content)
        session_nodes = xml_obj.getElementsByTagName("sessionKey")
        if not session_nodes:
            raise CredException("Invalid username or password.")
        session_key = session_nodes[0].firstChild.nodeValue
        if not session_key:
            raise CredException("Get session key failed.")
        return session_key

    def update(self, stanza):
        """
        Update or Create credentials based on the stanza
        :stanza: nested dict object. The outlayer keys are stanza name, and
                 inner dict is user/pass key/value pair to be encrypted
         {
         "stanza_name": {"tommy": "tommypasswod", "jerry": "jerrypassword"}
         }
        :return: raise on failure
        """

        for name, encr_dict in list(stanza.items()):
            encrypts = []
            for key, val in list(encr_dict.items()):
                encrypts.append(key)
                encrypts.append(val)
            self._update(name, self._sep.join(encrypts))

    def _update(self, name, str_to_encrypt):
        """
        Update the string for the name.
        :return: raise on failure
        """

        if len(str_to_encrypt) <= SPLUNK_CRED_LEN_LIMIT:
            self._do_update(name, str_to_encrypt)
            return

        # split the str_to_encrypt when len > 255
        length = SPLUNK_CRED_LEN_LIMIT
        i = 0
        while length < len(str_to_encrypt) + SPLUNK_CRED_LEN_LIMIT:
            curr_str = str_to_encrypt[length - SPLUNK_CRED_LEN_LIMIT : length]
            length += SPLUNK_CRED_LEN_LIMIT

            stanza_name = self._sep.join((name, str(i)))
            self._do_update(stanza_name, curr_str)
            i += 1

    def _do_update(self, name, password):
        try:
            self._create(name, password)
        except CredException:
            payload = {"password": password}
            endpoint = self._get_endpoint(name)
            response, _ = rest.splunkd_request(
                endpoint, self._session_key, method="POST", data=payload
            )
            if not response or response.status not in (200, 201):
                raise CredException(
                    "Unable to update password for username={}, status={}".format(
                        name, response.status
                    )
                )

    def _create(self, name, str_to_encrypt):
        """
        Create a new stored credential.
        :return: raise on failure
        """

        payload = {
            "name": name,
            "password": str_to_encrypt,
            "realm": self._realm,
        }

        endpoint = self._get_endpoint(name)
        resp, content = rest.splunkd_request(
            endpoint, self._session_key, method="POST", data=payload
        )
        if not resp or resp.status not in (200, 201, "200", "201"):
            raise CredException("Failed to encrypt username {}".format(name))

    def delete(self, name, throw=False):
        """
        Delete the encrypted entry
        """

        try:
            self._delete(name, throw=True)
        except CredNotFound:
            # try to delete the split stanzas
            try:
                stanzas = self._get_all_passwords()
            except Exception:
                raise

            prefix = self._realm + ":" + name + self._sep
            for stanza in stanzas:
                stanza_name = stanza.get("name")
                match = True
                try:
                    if stanza_name[: len(prefix)] != prefix:
                        match = False
                    num = stanza_name[len(prefix) : -1]
                    int(num)
                except (IndexError, ValueError):
                    match = False
                if match:
                    try:
                        delete_name = name + self._sep + num
                        self._delete(delete_name, throw=True)
                    except CredNotFound:
                        pass
                    except CredException:
                        raise
        except CredException:
            raise

    def _delete(self, name, throw=False):
        """
        Delete the encrypted entry
        """

        endpoint = self._get_endpoint(name)
        response, content = rest.splunkd_request(
            endpoint, self._session_key, method="DELETE"
        )

        if response is not None and response.status in (404, "404"):
            if throw:
                raise CredNotFound("Credential stanza not exits - {}".format(name))
        elif not response or response.status not in (200, 201, "200", "201"):
            if throw:
                raise CredException(
                    "Failed to delete credential stanza {}".format(name)
                )

    def get_all_passwords(self):
        results = {}
        all_stanzas = self._get_all_passwords()
        for stanza in all_stanzas:
            name = stanza.get("name")
            match = re.match(r"(.+){}(\d+)".format(self._sep), name)
            if match:
                actual_name = match.group(1) + ":"
                index = int(match.group(2))
                if results.get(actual_name):
                    exist_stanza = results.get(actual_name)
                else:
                    exist_stanza = stanza
                    exist_stanza["name"] = actual_name
                    exist_stanza["username"] = exist_stanza["username"].split(
                        self._sep
                    )[0]
                    exist_stanza["clears"] = {}
                    exist_stanza["encrs"] = {}

                try:
                    exist_stanza["clears"][index] = stanza.get("clear_password")
                    exist_stanza["encrs"][index] = stanza.get("encr_password")
                except KeyError:
                    exist_stanza["clears"] = {}
                    exist_stanza["encrs"] = {}
                    exist_stanza["clears"][index] = stanza.get("clear_password")
                    exist_stanza["encrs"][index] = stanza.get("encr_password")

                results[actual_name] = exist_stanza

            else:
                results[name] = stanza

        # merge the stanzas by index
        for name, stanza in list(results.items()):
            field_clear = stanza.get("clears")
            field_encr = stanza.get("encrs")
            if isinstance(field_clear, dict):
                clear_password = ""
                encr_password = ""
                for index in sorted(field_clear.keys()):
                    clear_password += field_clear.get(index)
                    encr_password += field_encr.get(index)
                stanza["clear_password"] = clear_password
                stanza["encr_password"] = encr_password

                del stanza["clears"]
                del stanza["encrs"]
        return list(results.values())

    def _get_all_passwords(self):
        """
        :return: a list of dict when successful, None when failed.
        the dict at least contains
        {
            "realm": xxx,
            "username": yyy,
            "clear_password": zzz,
        }
        """

        endpoint = self._get_endpoint()
        response, content = rest.splunkd_request(
            endpoint, self._session_key, method="GET"
        )
        if response and response.status in (200, 201, "200", "201") and content:
            return xdp.parse_conf_xml_dom(content)
        raise CredException("Failed to get credentials")

    def get_clear_password(self, name=None):
        """
        :return: clear password(s)
        {
        stanza_name: {"user": pass}
        }
        """

        return self._get_credentials("clear_password", name)

    def get_encrypted_password(self, name=None):
        """
        :return: encyrpted password(s)
        """

        return self._get_credentials("encr_password", name)

    def _get_credentials(self, prop, name=None):
        """
        :return: clear or encrypted password for specified realm, user
        """

        all_stanzas = self.get_all_passwords()
        results = {}

        for stanza in all_stanzas:
            if name and not stanza.get("name").endswith(":" + name + ":"):
                continue
            if stanza.get("realm") == self._realm:
                values = stanza[prop].split(self._sep)
                if len(values) % 2 == 1:
                    continue
                result = {values[i]: values[i + 1] for i in range(0, len(values), 2)}
                results[stanza.get("username")] = result
        return results

    @staticmethod
    def _build_name(realm, name):
        return util.format_stanza_name(
            "".join(
                (
                    CredentialManager._escape_string(realm),
                    ":",
                    CredentialManager._escape_string(name),
                    ":",
                )
            )
        )

    @staticmethod
    def _escape_string(string_to_escape):
        r"""
        Splunk secure credential storage actually requires a custom style of
        escaped string where all the :'s are escaped by a single \.
        But don't escape the control : in the stanza name.
        """

        return string_to_escape.replace(":", "\\:")

    def _get_endpoint(self, name=None, query=False):
        app = self._app
        owner = self._owner
        if query:
            app = "-"
            owner = "-"

        if name:
            realm_user = self._build_name(self._realm, name)
            rest_endpoint = "{}/servicesNS/{}/{}/storage/passwords/{}".format(
                self._splunkd_uri, owner, app, realm_user
            )
        else:
            rest_endpoint = "{}/servicesNS/{}/{}/storage/passwords?count=-1" "".format(
                self._splunkd_uri, owner, app
            )
        return rest_endpoint
